local addonName, addon = ...

--[[
Name: RatingBuster
Revision: $Revision: 78903 $
Author: Whitetooth, raethkcj
Description: Converts combat ratings in tooltips into normal percentages.
]]

---------------
-- Libraries --
---------------
local StatLogic = LibStub("StatLogic")
local GSM = function(...)
	return StatLogic:GetStatMod(...)
end
local L = LibStub("AceLocale-3.0"):GetLocale("RatingBuster")
---@cast L RatingBusterLocale

--------------------
-- AceAddon Setup --
--------------------
---@class RatingBuster: AceAddon, AceConsole-3.0, AceEvent-3.0, AceBucket-3.0
RatingBuster = LibStub("AceAddon-3.0"):NewAddon("RatingBuster", "AceConsole-3.0", "AceEvent-3.0", "AceBucket-3.0")
RatingBuster.title = "Rating Buster"
--@non-debug@
RatingBuster.version = "1.13.7"
--@end-non-debug@
--[==[@debug@
RatingBuster.version = "(development)"
--@end-debug@]==]
local addonNameWithVersion = ("%s %s"):format(addonName, RatingBuster.version)
RatingBuster.date = ("$Date: 2008-07-22 15:35:19 +0800 (星期二, 22 七月 2008) $"):gsub("^.-(%d%d%d%d%-%d%d%-%d%d).-$", "%1")

-----------
-- Cache --
-----------
local cache = {}
setmetatable(cache, {__mode = "kv"}) -- weak table to enable garbage collection
local function clearCache()
	for k in pairs(cache) do
		cache[k] = nil
	end
end
--debug
--RatingBuster.cache = cache

---------------------
-- Local Variables --
---------------------
local _
local _, class = UnitClass("player")
local calcLevel, playerLevel
local profileDB, globalDB -- Initialized in :OnInitialize()

-- Localize globals
local pairs = pairs
local ipairs = ipairs
local type = type
local select = select
local tinsert = tinsert
local tremove = tremove
local tsort = table.sort
local unpack = unpack
local tonumber = tonumber

local GetItemInfoCached = setmetatable({}, { __index = function(self, n)
	self[n] = {GetItemInfo(n)} -- store in cache
	return self[n] -- return result
end })
local GetItemInfo = function(item)
	return unpack(GetItemInfoCached[item])
end
local GetParryChance = GetParryChance
local GetBlockChance = GetBlockChance

---------------------------
-- Slash Command Options --
---------------------------

local function getOption(info)
	if type(globalDB[info[#info]]) ~= "nil" then
		return globalDB[info[#info]]
	else
		return profileDB[info[#info]]
	end
end
local function setOptionAndClearCache(info, value)
	if type(globalDB[info[#info]]) ~= "nil" then
		globalDB[info[#info]] = value
	else
		profileDB[info[#info]] = value
	end
	clearCache()
end
local function getGem(info)
	return profileDB[info[#info]].gemLink
end
local function setGem(info, value)
	if value == "" then
		profileDB[info[#info]].itemID = nil
		profileDB[info[#info]].gemID = nil
		profileDB[info[#info]].gemName = nil
		profileDB[info[#info]].gemLink = nil
		return
	end
	local gemID, gemText = StatLogic:GetGemID(value)
	if gemID and gemText then
		local name, link = GetItemInfo(value)
		local itemID = link:match("item:(%d+)")
		profileDB[info[#info]].itemID = itemID
		profileDB[info[#info]].gemID = gemID
		profileDB[info[#info]].gemName = name
		profileDB[info[#info]].gemLink = link
		-- Trim spaces
		gemText = gemText:trim()
		-- Strip color codes
		if gemText:sub(-2) == "|r" then
			gemText = gemText:sub(1, -3)
		end
		if gemText:sub(1, 10):find("|c%x%x%x%x%x%x%x%x") then
			gemText = gemText:sub(11)
		end
		profileDB[info[#info]].gemText = gemText
		clearCache()
		local socket = "EMPTY_SOCKET_" .. info[#info]:sub(7):upper()
		if not debugstack():find("AceConsole") then
			RatingBuster:Print(L["%s is now set to %s"]:format(_G[socket], link))
		end
	else
		RatingBuster:Print(L["Queried server for Gem: %s. Try again in 5 secs."]:format(value))
	end
end
local function getColor(info)
	local color = globalDB[info[#info]]
	if not color then
		color = profileDB[info[#info]]
	end
	return color:GetRGB()
end
local function setColor(info, r, g, b)
	local color = globalDB[info[#info]]
	if not color then
		color = profileDB[info[#info]]
	end
	color:SetRGB(r, g, b)
	clearCache()
end

ColorPickerFrame:SetMovable(true)
ColorPickerFrame:EnableMouse(true)
ColorPickerFrame:RegisterForDrag("LeftButton")
ColorPickerFrame:SetScript("OnDragStart", function()
	for _, frame in ipairs(C_System.GetFrameStack()) do
		if frame == ColorPickerFrameHeader then
			ColorPickerFrame:StartMoving()
		end
	end
end)
ColorPickerFrame:SetScript("OnDragStop", ColorPickerFrame.StopMovingOrSizing)

local options = {
	type = 'group',
	get = getOption,
	set = setOptionAndClearCache,
	args = {
		help = {
			type = 'execute',
			name = L["Help"],
			desc = L["Show this help message"],
			func = function()
				LibStub("AceConfigCmd-3.0").HandleCommand(RatingBuster, "rb", addonNameWithVersion, "")
			end
		},
		pp = {
			type = "execute",
			name = "Performance Profile",
			desc = "Execute a performance test and display the results",
			func = function()
				RatingBuster:PerformanceProfile()
			end,
			hidden = true,
		},
		enableStatMods = {
			type = 'toggle',
			name = L["Enable Stat Mods"],
			desc = L["Enable support for Stat Mods"],
		},
		showItemID = {
			type = 'toggle',
			name = L["Show ItemID"],
			desc = L["Show the ItemID in tooltips"],
		},
		showItemLevel = {
			type = 'toggle',
			name = L["Show ItemLevel"],
			desc = L["Show the ItemLevel in tooltips"],
		},
		useRequiredLevel = {
			type = 'toggle',
			name = L["Use required level"],
			desc = L["Calculate using the required level if you are below the required level"],
		},
		customLevel = {
			type = 'range',
			name = L["Set level"],
			desc = L["Set the level used in calculations (0 = your level)"],
			min = 0,
			max = GetMaxPlayerLevel(),
			step = 1,
		},
		rating = {
			type = 'group',
			name = L["Rating"],
			desc = L["Options for Rating display"],
			order = 1,
			args = {
				showRatings = {
					type = 'toggle',
					name = L["Show Rating conversions"],
					desc = L["Show Rating conversions in tooltips"],
				},
				ratingSpell = {
					type = 'toggle',
					name = L["Show Spell Hit/Haste"],
					desc = L["Show Spell Hit/Haste from Hit/Haste Rating"],
					hidden = function()
						local genericHit = StatLogic.GenericStatMap[StatLogic.Stats.HitRating]
						return (not genericHit) or (not tContains(genericHit, StatLogic.Stats.SpellHitRating))
					end
				},
				ratingPhysical = {
					type = 'toggle',
					name = L["Show Physical Hit/Haste"],
					desc = L["Show Physical Hit/Haste from Hit/Haste Rating"],
				},
				detailedConversionText = {
					type = 'toggle',
					name = L["Show detailed conversions text"],
					desc = L["Show detailed text for Resilience and Expertise conversions"],
				},
				defBreakDown = {
					type = 'toggle',
					name = L["Defense breakdown"],
					desc = L["Convert Defense into Crit Avoidance, Hit Avoidance, Dodge, Parry and Block"],
				},
				wpnBreakDown = {
					type = 'toggle',
					name = L["Weapon Skill breakdown"],
					desc = L["Convert Weapon Skill into Crit, Hit, Dodge Reduction, Parry Reduction and Block Reduction"],
				},
				expBreakDown = {
					type = 'toggle',
					name = L["Expertise breakdown"],
					desc = L["Convert Expertise into Dodge Reduction and Parry Reduction"],
				},
				enableAvoidanceDiminishingReturns = {
					type = 'toggle',
					name = L["Enable Avoidance Diminishing Returns"],
					desc = L["Dodge, Parry, Miss Avoidance values will be calculated using the avoidance deminishing return formula with your current stats"],
					hidden = function()
						return not StatLogic.GetAvoidanceAfterDR
					end,
				},
			},
		},
		stat = {
			type = 'group',
			name = L["Stat Breakdown"],
			desc = L["Changes the display of base stats"],
			order = 2,
			args = {
				showStats = {
					type = 'toggle',
					name = L["Show base stat conversions"],
					desc = L["Show base stat conversions in tooltips"],
					width = "full",
					order = 1,
				},
				textColor = {
					type = 'color',
					name = L["Change text color"],
					desc = L["Changes the color of added text"],
					get = getColor,
					set = setColor,
					order = 2,
				},
				str = {
					type = 'group',
					name = L[StatLogic.Stats.Strength],
					desc = L["Changes the display of %s"]:format(L[StatLogic.Stats.Strength]),
					width = "full",
					order = 3,
					args = {},
				},
				agi = {
					type = 'group',
					name = L[StatLogic.Stats.Agility],
					desc = L["Changes the display of %s"]:format(L[StatLogic.Stats.Agility]),
					width = "full",
					order = 4,
					args = {
						showCritFromAgi = {
							type = 'toggle',
							name = L["Show %s"]:format(L[StatLogic.Stats.MeleeCrit]),
							width = "full",
						},
						showDodgeFromAgi = {
							type = 'toggle',
							name = L["Show %s"]:format(L[StatLogic.Stats.Dodge]),
							width = "full",
						},
					},
				},
				sta = {
					type = 'group',
					name = L[StatLogic.Stats.Stamina],
					desc = L["Changes the display of %s"]:format(L[StatLogic.Stats.Stamina]),
					width = "full",
					order = 5,
					args = {},
				},
				int = {
					type = 'group',
					name = L[StatLogic.Stats.Intellect],
					desc = L["Changes the display of %s"]:format(L[StatLogic.Stats.Intellect]),
					width = "full",
					order = 6,
					args = {
						showSpellCritFromInt = {
							type = 'toggle',
							name = L["Show %s"]:format(L[StatLogic.Stats.SpellCrit]),
							width = "full",
						},
					},
				},
				spi = {
					type = 'group',
					name = L[StatLogic.Stats.Spirit],
					desc = L["Changes the display of %s"]:format(L[StatLogic.Stats.Spirit]),
					width = "full",
					order = 7,
					args = {},
				},
				ap = {
					type = 'group',
					name = L[StatLogic.Stats.AttackPower],
					desc = L["Changes the display of %s"]:format(L[StatLogic.Stats.AttackPower]),
					order = 8,
					args = {},
					hidden = true,
				},
				armor = {
					type = 'group',
					name = L[StatLogic.Stats.Armor],
					desc = L["Changes the display of %s"]:format(L[StatLogic.Stats.Armor]),
					order = 9,
					args = {},
					hidden = true,
				},
			},
		},
		sum = {
			type = 'group',
			name = L["Stat Summary"],
			desc = L["Options for stat summary"],
			order = 3,
			args = {
				showSum = {
					type = 'toggle',
					name = L["Show stat summary"],
					desc = L["Show stat summary in tooltips"],
					order = 1,
				},
				calcSum = {
					type = 'toggle',
					name = L["Calculate stat sum"],
					desc = L["Calculate the total stats for the item"],
					order = 2,
				},
				calcDiff = {
					type = 'toggle',
					name = L["Calculate stat diff"],
					desc = L["Calculate the stat difference for the item and equipped items"],
					order = 3,
				},
				sumDiffStyle = {
					type = 'select',
					name = L["Display style for diff value"],
					desc = L["Display diff values in the main tooltip or only in compare tooltips"],
					values = {
						["comp"] = "Compare",
						["main"] = "Main"
					},
					order = 4,
				},
				hideBlizzardComparisons = {
					type = 'toggle',
					name = L["Hide Blizzard Item Comparisons"],
					desc = L["Disable Blizzard stat change summary when using the built-in comparison tooltip"],
					width = "double",
					order = 4.5,
				},
				sumShowIcon = {
					type = 'toggle',
					name = L["Show icon"],
					desc = L["Show the sigma icon before summary listing"],
					order = 5,
				},
				sumShowTitle = {
					type = 'toggle',
					name = L["Show title text"],
					desc = L["Show the title text before summary listing"],
					order = 6,
				},
				showZeroValueStat = {
					type = 'toggle',
					name = L["Show zero value stats"],
					desc = L["Show zero value stats in summary for consistancy"],
					order = 7,
				},
				sumSortAlpha = {
					type = 'toggle',
					name = L["Sort StatSummary alphabetically"],
					desc = L["Enable to sort StatSummary alphabetically, disable to sort according to stat type(basic, physical, spell, tank)"],
					order = 8,
				},
				space = {
					type = 'group',
					name = L["Add empty line"],
					inline = true,
					args = {
						sumBlankLine = {
							type = 'toggle',
							name = L["Add before summary"],
							desc = L["Add a empty line before stat summary"],
							order = 10,
						},
						sumBlankLineAfter = {
							type = 'toggle',
							name = L["Add after summary"],
							desc = L["Add a empty line after stat summary"],
							order = 11,
						},
					},
				},
				sumStatColor = {
					type = 'color',
					name = L["Change text color"],
					desc = L["Changes the color of added text"],
					get = getColor,
					set = setColor,
					order = 13,
				},
				sumValueColor = {
					type = 'color',
					name = L["Change number color"],
					desc = L["Changes the color of added text"],
					get = getColor,
					set = setColor,
					order = 14,
				},
				ignore = {
					type = 'group',
					name = L["Ignore settings"],
					desc = L["Ignore stuff when calculating the stat summary"],
					args = {
						sumIgnoreUnused = {
							type = 'toggle',
							name = L["Ignore unused item types"],
							desc = L["Show stat summary only for highest level armor type and items you can use with uncommon quality and up"],
						},
						sumIgnoreEquipped = {
							type = 'toggle',
							name = L["Ignore equipped items"],
							desc = L["Hide stat summary for equipped items"],
						},
						sumIgnoreEnchant = {
							type = 'toggle',
							name = L["Ignore enchants"],
							desc = L["Ignore enchants on items when calculating the stat summary"],
						},
						sumIgnoreGems = {
							type = 'toggle',
							name = L["Ignore gems"],
							desc = L["Ignore gems on items when calculating the stat summary"],
						},
						sumIgnoreExtraSockets = {
							type = 'toggle',
							name = L["Ignore extra sockets"],
							desc = L["Ignore sockets from professions or consumable items when calculating the stat summary"]
						}
					},
				},
				basic = {
					type = 'group',
					name = L["Stat - Basic"],
					desc = L["Choose basic stats for summary"],
					args = {
						sumHP = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.Health]),
							desc = L["Health <- Health, Stamina"],
							order = 1,
						},
						sumMP = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.Mana]),
							desc = L["Mana <- Mana, Intellect"],
							order = 2,
						},
						sumMP5 = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.ManaRegen]),
							desc = L["Mana Regen <- Mana Regen, Spirit"],
							order = 3,
						},
						sumMP5NC = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.ManaRegenNotCasting]),
							desc = L["Mana Regen while not casting <- Spirit"],
							order = 4,
						},
						sumHP5 = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.HealthRegen]),
							desc = L["Health Regen <- Health Regen"],
							order = 5,
						},
						sumHP5OC = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.HealthRegenOutOfCombat]),
							desc = L["Health Regen when out of combat <- Spirit"],
							order = 6,
						},
						sumStr = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.Strength]),
							order = 7,
						},
						sumAgi = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.Agility]),
							order = 8,
						},
						sumSta = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.Stamina]),
							order = 9,
						},
						sumInt = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.Intellect]),
							order = 10,
						},
						sumSpi = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.Spirit]),
							order = 11,
						},
					},
				},
				physical = {
					type = 'group',
					name = L["Stat - Physical"],
					desc = L["Choose physical damage stats for summary"],
					args = {
						sumAP = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.AttackPower]),
							desc = L["Attack Power <- Attack Power, Strength, Agility"],
							width = "double",
							order = 1,
						},
						sumHit = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.MeleeHit]),
							desc = L["Hit Chance <- Hit Rating, Weapon Skill Rating"],
							order = 2,
						},
						sumHitRating = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.MeleeHitRating]),
							order = 3,
						},
						sumCrit = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.MeleeCrit]),
							desc = L["Crit Chance <- Crit Rating, Agility, Weapon Skill Rating"],
							order = 4,
						},
						sumCritRating = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.MeleeCritRating]),
							order = 5,
						},
						sumHaste = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.MeleeHaste]),
							desc = L["Haste <- Haste Rating"],
							order = 6,
						},
						sumHasteRating = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.MeleeHasteRating]),
							order = 7,
						},
						sumIgnoreArmor = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.IgnoreArmor]),
							hidden = function()
								return StatLogic:RatingExists(StatLogic.Stats.ArmorPenetrationRating)
							end,
							order = 8,
						},
						sumArmorPenetration = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.ArmorPenetration]),
							hidden = function()
								return not StatLogic:RatingExists(StatLogic.Stats.ArmorPenetrationRating)
							end,
							order = 9,
						},
						sumArmorPenetrationRating = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.ArmorPenetrationRating]),
							hidden = function()
								return not StatLogic:RatingExists(StatLogic.Stats.ArmorPenetrationRating)
							end,
							order = 10,
						},
						ranged = {
							type = 'header',
							name = L["Ranged"],
							order = 11,
						},
						sumRAP = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.RangedAttackPower]),
							desc = L["Ranged Attack Power <- Ranged Attack Power, Intellect, Attack Power, Strength, Agility"],
							width = "double",
							order = 12,
						},
						sumRangedHit = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.RangedHit]),
							desc = L["Ranged Hit Chance <- Hit Rating, Weapon Skill Rating, Ranged Hit Rating"],
							order = 13,
						},
						sumRangedHitRating = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.RangedHitRating]),
							order = 14,
						},
						sumRangedCrit = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.RangedCrit]),
							desc = L["Ranged Crit Chance <- Crit Rating, Agility, Weapon Skill Rating, Ranged Crit Rating"],
							order = 15,
						},
						sumRangedCritRating = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.RangedCritRating]),
							order = 16,
						},
						sumRangedHaste = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.RangedHaste]),
							desc = L["Ranged Haste <- Haste Rating, Ranged Haste Rating"],
							order = 17,
						},
						sumRangedHasteRating = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.RangedHasteRating]),
							order = 18,
						},
						weapon = {
							type = 'header',
							name = L["Weapon"],
							order = 19,
						},
						sumWeaponAverageDamage = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.AverageWeaponDamage]),
							order = 20,
						},
						sumWeaponDPS = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.WeaponDPS]),
							order = 21,
						},
						sumDodgeNeglect = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.DodgeReduction]),
							desc = L["Dodge Reduction <- Expertise, Weapon Skill Rating"],
							order = 22,
						},
						sumParryNeglect = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.ParryReduction]),
							desc = L["Parry Reduction <- Expertise, Weapon Skill Rating"],
							order = 23,
						},
						sumWeaponSkill = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.WeaponSkill]),
							desc = L["Weapon Skill <- Weapon Skill Rating"],
							order = 24,
						},
						sumExpertise = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.Expertise]),
							desc = L["Expertise <- Expertise Rating"],
							order = 25,
						},
					},
				},
				spell = {
					type = 'group',
					name = L["Stat - Spell"],
					desc = L["Choose spell damage and healing stats for summary"],
					args = {
						sumSpellDmg = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.SpellDamage]),
							desc = L["Spell Damage <- Spell Damage, Intellect, Spirit, Stamina"],
						},
						sumHolyDmg = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.HolyDamage]),
							desc = L["Holy Spell Damage <- Holy Spell Damage, Spell Damage, Intellect, Spirit"],
						},
						sumArcaneDmg = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.ArcaneDamage]),
							desc = L["Arcane Spell Damage <- Arcane Spell Damage, Spell Damage, Intellect"],
						},
						sumFireDmg = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.FireDamage]),
							desc = L["Fire Spell Damage <- Fire Spell Damage, Spell Damage, Intellect, Stamina"],
						},
						sumNatureDmg = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.NatureDamage]),
							desc = L["Nature Spell Damage <- Nature Spell Damage, Spell Damage, Intellect"],
						},
						sumFrostDmg = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.FrostDamage]),
							desc = L["Frost Spell Damage <- Frost Spell Damage, Spell Damage, Intellect"],
						},
						sumShadowDmg = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.ShadowDamage]),
							desc = L["Shadow Spell Damage <- Shadow Spell Damage, Spell Damage, Intellect, Spirit, Stamina"],
						},
						sumHealing = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.HealingPower]),
							desc = L["Healing <- Healing, Intellect, Spirit, Agility, Strength"],
						},
						sumSpellHit = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.SpellHit]),
							desc = L["Spell Hit Chance <- Spell Hit Rating"],
						},
						sumSpellHitRating = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.SpellHitRating]),
						},
						sumSpellCrit = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.SpellCrit]),
							desc = L["Spell Crit Chance <- Spell Crit Rating, Intellect"],
						},
						sumSpellCritRating = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.SpellCritRating]),
						},
						sumSpellHaste = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.SpellHaste]),
							desc = L["Spell Haste <- Spell Haste Rating"],
						},
						sumSpellHasteRating = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.SpellHasteRating]),
						},
						sumPenetration = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.SpellPenetration]),
						},
					},
				},
				tank = {
					type = 'group',
					name = L["Stat - Tank"],
					desc = L["Choose tank stats for summary"],
					args = {
						sumAvoidance = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.Avoidance]),
							desc = L["Avoidance <- Dodge, Parry, MobMiss, Block(Optional)"],
							order = 1,
						},
						sumAvoidWithBlock = {
							type = 'toggle',
							name = L["Include block chance in Avoidance summary"],
							desc = L["Enable to include block chance in Avoidance summary, Disable for only dodge, parry, miss"],
							order = 2,
						},
						sumDodge = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.Dodge]),
							desc = L["Dodge Chance <- Dodge Rating, Agility, Defense Rating"],
							order = 3,
						},
						sumDodgeRating = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.DodgeRating]),
							order = 4,
						},
						sumBlock = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.BlockChance]),
							desc = L["Block Chance <- Block Rating, Defense Rating"],
							order = 5,
						},
						sumBlockRating = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.BlockRating]),
							order = 6,
						},
						sumBlockValue = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.BlockValue]),
							desc = L["Block Value <- Block Value, Strength"],
							order = 7,
						},
						sumParry = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.Parry]),
							desc = L["Parry Chance <- Parry Rating, Defense Rating"],
							order = 8,
						},
						sumParryRating = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.ParryRating]),
							order = 9,
						},
						sumHitAvoid = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.Miss]),
							desc = L["Hit Avoidance <- Defense Rating"],
							order = 10,
						},
						sumArmor = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.Armor]),
							desc = L["Armor <- Armor from items, Armor from bonuses, Agility, Intellect"],
							order = 11,
						},
						sumDefense = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.Defense]),
							desc = L["Defense <- Defense Rating"],
							order = 12,
						},
						sumCritAvoid = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.CritAvoidance]),
							desc = L["Crit Avoidance <- Defense Rating, Resilience"],
							order = 13,
						},
						sumResilience = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.ResilienceRating]),
							order = 14,
						},
						sumArcaneResist = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.ArcaneResistance]),
							order = 15,
						},
						sumFireResist = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.FireResistance]),
							order = 16,
						},
						sumNatureResist = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.NatureResistance]),
							order = 17,
						},
						sumFrostResist = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.FrostResistance]),
							order = 18,
						},
						sumShadowResist = {
							type = 'toggle',
							name = L["Sum %s"]:format(L[StatLogic.Stats.ShadowResistance]),
							order = 19,
						},
					},
				},
				gem = {
					type = 'group',
					name = L["Gems"],
					desc = L["Auto fill empty gem slots"],
					args = {
						sumGemRed = {
							type = 'input',
							name = EMPTY_SOCKET_RED,
							desc = L["ItemID or Link of the gem you would like to auto fill"],
							usage = L["<ItemID|Link>"],
							get = getGem,
							set = setGem,
							order = 1,
						},
						sumGemYellow = {
							type = 'input',
							name = EMPTY_SOCKET_YELLOW,
							desc = L["ItemID or Link of the gem you would like to auto fill"],
							usage = L["<ItemID|Link>"],
							get = getGem,
							set = setGem,
							order = 2,
						},
						sumGemBlue = {
							type = 'input',
							name = EMPTY_SOCKET_BLUE,
							desc = L["ItemID or Link of the gem you would like to auto fill"],
							usage = L["<ItemID|Link>"],
							get = getGem,
							set = setGem,
							order = 3,
						},
						sumGemMeta = {
							type = 'input',
							name = EMPTY_SOCKET_META,
							desc = L["ItemID or Link of the gem you would like to auto fill"],
							usage = L["<ItemID|Link>"],
							get = getGem,
							set = setGem,
							order = 4,
						},
					},
				},
			},
		},
		alwaysBuffed = {
			type = "group",
			name = "AlwaysBuffed",
			order = 4,
			get = function(info)
				local db = RatingBuster.db:GetNamespace("AlwaysBuffed")
				return db.profile[info[#info]]
			end,
			set = function(info, v)
				local db = RatingBuster.db:GetNamespace("AlwaysBuffed")
				db.profile[info[#info]] = v
				clearCache()
				StatLogic:InvalidateEvent("UNIT_AURA", "player")
			end,
      args = {
        description = {
					type = "description",
          name = L["Enables RatingBuster to calculate selected buff effects even if you don't really have them"],
          order = 1,
        },
        description2 = {
					type = "description",
          name = " ",
          order = 2,
        },
        [class] = {
					type = "group",
					dialogInline = true,
          name = L["$class Self Buffs"]:gsub("$class", (UnitClass("player"))),
          order = 5,
					hidden = true,
					args = {},
        },
        ALL = {
					type = "group",
					dialogInline = true,
          name = L["Raid Buffs"],
          order = 6,
					args = {},
        },
      },
		},
	},
}

---------------------
-- Saved Variables --
---------------------
-- Default values
local defaults = {
	global = {
		showItemLevel = false,
		showItemID = false,
		useRequiredLevel = true,
		customLevel = 0,
		textColor = CreateColor(1.0, 0.996, 0.545),
		showSum = true,
		sumIgnoreUnused = true,
		sumIgnoreEquipped = false,
		sumIgnoreEnchant = true,
		sumIgnoreGems = false,
		sumIgnoreExtraSockets = true,
		sumBlankLine = true,
		sumBlankLineAfter = false,
		sumStatColor = CreateColor(NORMAL_FONT_COLOR:GetRGBA()),
		sumValueColor = CreateColor(NORMAL_FONT_COLOR:GetRGBA()),
		sumShowIcon = true,
		sumShowTitle = true,
		sumDiffStyle = "main",
		sumSortAlpha = false,
		calcDiff = true,
		calcSum = true,
		hideBlizzardComparisons = true,
	},
	profile = {
		enableStatMods = true,
		enableAvoidanceDiminishingReturns = StatLogic.GetAvoidanceAfterDR and true or false,
		showRatings = true,
		detailedConversionText = false,
		defBreakDown = false,
		wpnBreakDown = false,
		expBreakDown = false,
		showStats = true,
		sumAvoidWithBlock = false,
		showZeroValueStat = false,
		--[[
		Str -> AP, Block
		Agi -> Crit, Dodge, AP, RAP, Armor
		Sta -> Health
		Int -> Mana, SpellCrit, MP5NC
		Spi -> MP5NC, HP5
		--]]
		-- Base stat conversions
		showAPFromStr = false,
		showBlockValueFromStr = false,

		showCritFromAgi = true,
		showDodgeFromAgi = true,
		showAPFromAgi = false,
		showRAPFromAgi = false,
		showArmorFromAgi = false,

		showHealthFromSta = false,

		showManaFromInt = false,
		showSpellCritFromInt = true,
		showMP5NCFromInt = false,

		showMP5NCFromSpi = false,
		showHP5NCFromSpi = false,
		------------------
		-- Stat Summary --
		------------------
		-- Basic
		sumHP = true,
		sumMP = true,
		sumMP5 = true,
		sumMP5NC = false,
		sumHP5 = false,
		sumHP5OC = false,
		sumStr = false,
		sumAgi = false,
		sumSta = false,
		sumInt = false,
		sumSpi = false,
		-- Physical
		sumAP = false,
		sumHit = false,
		sumHitRating = false,
		sumCrit = false,
		sumCritRating = false,
		sumHaste = false,
		sumHasteRating = false,
		sumIgnoreArmor = false,
		sumArmorPenetration = false,
		-- Ranged
		sumRAP = false,
		sumRangedHit = false,
		sumRangedHitRating = false,
		sumRangedCrit = false,
		sumRangedCritRating = false,
		sumRangedHaste = false,
		sumRangedHasteRating = false,
		-- Weapon
		sumWeaponAverageDamage = false,
		sumWeaponDPS = false,
		sumExpertise = false,
		sumWeaponSkill = false,
		sumDodgeNeglect = false,
		sumParryNeglect = false,
		-- Spell
		sumSpellDmg = false,
		sumArcaneDmg = false,
		sumFrostDmg = false,
		sumNatureDmg = false,
		sumFireDmg = false,
		sumShadowDmg = false,
		sumHolyDmg = false,
		sumHealing = false,
		sumSpellHit = false,
		sumSpellHitRating = false, -- new
		sumSpellCrit = false,
		sumSpellCritRating = false, -- new
		sumSpellHaste = false, -- new
		sumSpellHasteRating = false, -- new
		sumPenetration = false, -- new
		-- Tank
		sumArmor = false,
		sumDodge = false,
		sumDodgeRating = false, -- new
		sumParry = false,
		sumParryRating = false, -- new
		sumBlock = false,
		sumBlockRating = false, -- new
		sumBlockValue = false,
		sumHitAvoid = false,
		sumCritAvoid = false,
		sumArcaneResist = false,
		sumFrostResist = false,
		sumNatureResist = false,
		sumFireResist = false,
		sumShadowResist = false,
		sumResilience = true, -- new
		sumDefense = false,
		sumAvoidance = false,
		-- Gems
		sumGemRed = {
			itemID = nil,
			gemID = nil,
			gemText = nil,
		};
		sumGemYellow = {
			itemID = nil,
			gemID = nil,
			gemText = nil,
		};
		sumGemBlue = {
			itemID = nil,
			gemID = nil,
			gemText = nil,
		};
		sumGemMeta = {
			itemID = nil,
			gemID = nil,
			gemText = nil,
		};
	}
}

-- Class specific settings
if class == "DEATHKNIGHT" then
	defaults.profile.sumWeaponAverageDamage = true
	defaults.profile.sumAvoidance = true
	defaults.profile.sumArmor = true
	defaults.profile.sumMP = false
	defaults.profile.sumMP5 = false
	defaults.profile.sumAP = true
	defaults.profile.sumHit = true
	defaults.profile.sumCrit = true
	defaults.profile.sumHaste = true
	defaults.profile.sumExpertise = true
	defaults.profile.showSpellCritFromInt = false
	defaults.profile.ratingPhysical = true
	defaults.profile.sumArmorPenetration = true
elseif class == "DRUID" then
	defaults.profile.sumAP = true
	defaults.profile.sumHit = true
	defaults.profile.sumCrit = true
	defaults.profile.sumHaste = true
	defaults.profile.sumExpertise = true
	defaults.profile.sumAvoidance = true
	defaults.profile.sumArmor = true
	defaults.profile.sumSpellDmg = true
	defaults.profile.sumSpellHit = true
	defaults.profile.sumSpellCrit = true
	defaults.profile.sumSpellHaste = true
	defaults.profile.sumHealing = true
	defaults.profile.ratingPhysical = true
	defaults.profile.ratingSpell = true
	defaults.profile.sumArmorPenetration = true
elseif class == "HUNTER" then
	defaults.profile.sumWeaponAverageDamage = true
	defaults.profile.sumRAP = true
	defaults.profile.sumRangedHit = true
	defaults.profile.sumRangedCrit = true
	defaults.profile.sumRangedHaste = true
	defaults.profile.showRAPFromAgi = true
	defaults.profile.showDodgeFromAgi = false
	defaults.profile.showSpellCritFromInt = false
	defaults.profile.showRAPFromInt = true
	defaults.profile.ratingPhysical = true
	defaults.profile.sumArmorPenetration = true
elseif class == "MAGE" then
	defaults.profile.sumSpellDmg = true
	defaults.profile.sumSpellHit = true
	defaults.profile.sumSpellCrit = true
	defaults.profile.sumSpellHaste = true
	defaults.profile.showCritFromAgi = false
	defaults.profile.showDodgeFromAgi = false
	defaults.profile.ratingSpell = true
elseif class == "PALADIN" then
	defaults.profile.sumWeaponAverageDamage = true
	defaults.profile.sumAvoidance = true
	defaults.profile.sumArmor = true
	defaults.profile.sumHit = true
	defaults.profile.sumCrit = true
	defaults.profile.sumHaste = true
	defaults.profile.sumExpertise = true
	defaults.profile.sumHolyDmg = true
	defaults.profile.sumSpellHit = true
	defaults.profile.sumSpellCrit = true
	defaults.profile.sumSpellHaste = true
	defaults.profile.sumHealing = true
	defaults.profile.ratingPhysical = true
	defaults.profile.ratingSpell = true
elseif class == "PRIEST" then
	defaults.profile.sumSpellDmg = true
	defaults.profile.sumSpellHit = true
	defaults.profile.sumSpellCrit = true
	defaults.profile.sumSpellHaste = true
	defaults.profile.sumHealing = true
	defaults.profile.showCritFromAgi = false
	defaults.profile.showDodgeFromAgi = false
	defaults.profile.ratingSpell = true
elseif class == "ROGUE" then
	defaults.profile.sumWeaponAverageDamage = true
	defaults.profile.sumMP = false
	defaults.profile.sumMP5 = false
	defaults.profile.sumAP = true
	defaults.profile.sumHit = true
	defaults.profile.sumCrit = true
	defaults.profile.sumHaste = true
	defaults.profile.sumExpertise = true
	defaults.profile.showSpellCritFromInt = false
	defaults.profile.ratingPhysical = true
	defaults.profile.sumArmorPenetration = true
elseif class == "SHAMAN" then
	defaults.profile.sumWeaponAverageDamage = true
	defaults.profile.sumSpellDmg = true
	defaults.profile.sumSpellHit = true
	defaults.profile.sumSpellCrit = true
	defaults.profile.sumSpellHaste = true
	defaults.profile.sumHealing = true
	defaults.profile.showDodgeFromAgi = false
	defaults.profile.ratingPhysical = true
	defaults.profile.ratingSpell = true
elseif class == "WARLOCK" then
	defaults.profile.sumSpellDmg = true
	defaults.profile.sumSpellHit = true
	defaults.profile.sumSpellCrit = true
	defaults.profile.sumSpellHaste = true
	defaults.profile.showCritFromAgi = false
	defaults.profile.showDodgeFromAgi = false
	defaults.profile.ratingSpell = true
elseif class == "WARRIOR" then
	defaults.profile.sumWeaponAverageDamage = true
	defaults.profile.sumAvoidance = true
	defaults.profile.sumArmor = true
	defaults.profile.sumMP = false
	defaults.profile.sumMP5 = false
	defaults.profile.sumAP = true
	defaults.profile.sumHit = true
	defaults.profile.sumCrit = true
	defaults.profile.sumHaste = true
	defaults.profile.sumExpertise = true
	defaults.profile.showSpellCritFromInt = false
	defaults.profile.ratingPhysical = true
	defaults.profile.sumArmorPenetration = true
end

-- Generate options from expansion-specific StatModTables in StatLogic
do
	-- Backwards compatibility
	local statToOptionKey = setmetatable({
		["AP"] = "AP",
		["MANA_REG"] = "MP5",
		["NORMAL_MANA_REG"] = "MP5NC",
		["HEALTH_REG"] = "HP5",
		["NORMAL_HEALTH_REG"] = "HP5NC",
		["RANGED_AP"] = "RAP",
	},
	{
		__index = function(_, stat)
			-- Remove underscores, PascalCase
			return (stat:gsub("[%W_]*(%w+)[%W_]*", function(word)
				return word:lower():gsub("^%l", string.upper)
			end))
		end
	})

	-- Backwards compatibility
	local statStringToStat = setmetatable({
		["AP"] = StatLogic.Stats.AttackPower,
		["RANGED_AP"] = StatLogic.Stats.RangedAttackPower,
		["MANA"] = StatLogic.Stats.Mana,
		["MANA_REG"] = StatLogic.Stats.ManaRegen,
		["NORMAL_MANA_REG"] = StatLogic.Stats.ManaRegenNotCasting,
		["HEALTH"] = StatLogic.Stats.Health,
		["HEALTH_REG"] = StatLogic.Stats.HealthRegen,
		["NORMAL_HEALTH_REG"] = StatLogic.Stats.HealthRegenOutOfCombat,
		["SPELL_DMG"] = StatLogic.Stats.SpellDamage,
		["SPELL_HIT"] = StatLogic.Stats.SpellHit,
		["SPELL_CRIT"] = StatLogic.Stats.SpellCrit,
		["HEALING"] = StatLogic.Stats.HealingPower,
		["PARRY"] = StatLogic.Stats.Parry,
		["BLOCK_VALUE"] = StatLogic.Stats.BlockValue,
	},
	{
		__index = function(_, stat)
			return stat
		end
	})

	local addStatModOption = function(add, mod, sources)
		-- ADD_HEALING_MOD_INT -> showHealingFromInt
		local key = "show" .. statToOptionKey[tostring(add)] .. "From" .. statToOptionKey[mod]
		if defaults.profile[key] == nil then
			defaults.profile[key] = true
		end

		-- Override Armor and Attack Power being hidden by default,
		-- since most classes have no useful breakdowns from them.
		local group = options.args.stat.args[mod:lower()]
		group.hidden = false

		local option = group.args[key]
		if not option then
			option = {
				type = "toggle",
				width = "full",
			}
		else
			sources = option.desc .. ", " .. sources
		end

		option.name = L["Show %s"]:format(L[statStringToStat[add]])
		option.desc = sources

		options.args.stat.args[mod:lower()].args[key] = option
	end

	local function GenerateStatModOptions()
		for _, modList in pairs(StatLogic.StatModTable) do
			for statMod, cases in pairs(modList) do
				local add = StatLogic.StatModInfo[statMod].add
				local mod = StatLogic.StatModInfo[statMod].mod

				if add and mod then
					local sources = ""
					local firstSource = true
					for _, case in ipairs(cases) do
						if not firstSource then
							sources = sources .. ", "
						end
						local source = ""
						if case.aura then
							source = GetSpellInfo(case.aura)
						elseif case.tab then
							source = StatLogic:GetOrderedTalentInfo(case.tab, case.num)
						elseif case.glyph then
							source = GetSpellInfo(case.glyph)
						elseif case.spellid then
							source = GetSpellInfo(case.spellid)
						end
						sources = sources .. source
						firstSource = false
					end

					-- Molten Armor and Forceful Deflection give rating,
					-- but we show it to the user as the converted stat
					add = add:gsub("_RATING", "")

					-- We want to show the user Armor, regardless of where it comes from
					if add == "BONUS_ARMOR" then
						add = StatLogic.Stats.Armor
					end

					if mod == "NORMAL_MANA_REG" then
						mod = "SPI"
						if GSM("ADD_NORMAL_MANA_REG_MOD_INT") > 0 then
							-- "Normal mana regen" is added from both int and spirit
							addStatModOption(add, "INT", sources)
						end
					elseif mod == "NORMAL_HEALTH_REG" then
						mod = "SPI"
					elseif mod == "MANA" then
						mod = "INT"
					elseif mod == "AP" then
						-- Paladin's Sheathe of Light, Touched by the Light
						-- Shaman's Mental Quickness.
						-- TODO: Shaman AP can also come from INT in Wrath
						if GSM("ADD_AP_MOD_STR") > 0 then
							addStatModOption(add, "STR", sources)
						end
						if GSM("ADD_AP_MOD_AGI") > 0 then
							addStatModOption(add, "AGI", sources)
						end
					end

					-- Demonic Knowledge technically scales with pet stats,
					-- but we compute the scaling from player's stats
					mod = mod:gsub("^PET_", "")

					addStatModOption(add, mod, sources)
				end
			end
		end
	end

	local season = C_Seasons and C_Seasons.HasActiveSeason() and C_Seasons.GetActiveSeason()
	local showRunes =  season ~= Enum.SeasonID.NoSeason and season ~= Enum.SeasonID.Hardcore

	local function GenerateAuraOptions()
		for modType, modList in pairs(StatLogic.StatModTable) do
			for modName, mods in pairs(modList) do
				if not StatLogic.StatModIgnoresAlwaysBuffed[modName] then
					for _, mod in ipairs(mods) do
						if mod.aura and (not mod.rune or showRunes) then
							local name, _, icon = GetSpellInfo(mod.aura)
							local option = {
								type = 'toggle',
								name = "|T"..icon..":25:25:-2:0|t"..name,
							}
							options.args.alwaysBuffed.args[modType].args[name] = option
							options.args.alwaysBuffed.args[modType].hidden = false

							-- If it's a spell the player knows, use the highest rank for the description
							local spellId = mod.aura
							if IsPlayerSpell(spellId) then
								spellId = select(7,GetSpellInfo(name)) or spellId
							end
							local spell = Spell:CreateFromSpellID(spellId)
							spell:ContinueOnSpellLoad(function()
								option.desc = spell:GetSpellDescription()
							end)
						end
					end
				end
			end
		end
	end

	local f = CreateFrame("Frame")
	f:RegisterEvent("SPELLS_CHANGED")
	f:SetScript("OnEvent", function()
		if StatLogic:TalentCacheExists() then
			GenerateStatModOptions()
			GenerateAuraOptions()
			RatingBuster:InitializeDatabase()
		else
			-- Talents are not guaranteed to exist on SPELLS_CHANGED,
			-- and there is no definite event for when they will exist.
			-- Recheck every 1 second after SPELLS_CHANGED until they exist.
			local ticker
			ticker = C_Timer.NewTicker(1, function()
				if StatLogic:TalentCacheExists() then
					GenerateStatModOptions()
					GenerateAuraOptions()
					RatingBuster:InitializeDatabase()
					ticker:Cancel()
				end
			end)
		end
		f:UnregisterEvent("SPELLS_CHANGED")
	end)
end

-----------
-- Tools --
-----------
-- copyTable
local function copyTable(to, from)
	if to then
		for k in pairs(to) do
			to[k] = nil
		end
		setmetatable(to, nil)
	else
		to = {}
	end
	for k,v in pairs(from) do
		if type(k) == "table" then
			k = copyTable({}, k)
		end
		if type(v) == "table" then
			v = copyTable({}, v)
		end
		to[k] = v
	end
	setmetatable(to, getmetatable(from))
	return to
end

---------------------
-- Initializations --
---------------------
--[[ Loading Process Event Reference
{
ADDON_LOADED - When this addon is loaded
VARIABLES_LOADED - When all addons are loaded
PLAYER_LOGIN - Most information about the game world should now be available to the UI
}
--]]
-- OnInitialize(name) called at ADDON_LOADED
function RatingBuster:RefreshConfig()
	profileDB = self.db.profile
	wipe(cache)
end

function RatingBuster:OnInitialize()
	LibStub("AceConfig-3.0"):RegisterOptionsTable(addonNameWithVersion, options)
	LibStub("AceConfigDialog-3.0"):AddToBlizOptions(addonNameWithVersion, addonName)
end

function RatingBuster:InitializeDatabase()
	RatingBuster.db = LibStub("AceDB-3.0"):New("RatingBusterDB", defaults, class)
	RatingBuster.db.RegisterCallback(RatingBuster, "OnProfileChanged", "RefreshConfig")
	RatingBuster.db.RegisterCallback(RatingBuster, "OnProfileCopied", "RefreshConfig")
	RatingBuster.db.RegisterCallback(RatingBuster, "OnProfileReset", "RefreshConfig")
	profileDB = RatingBuster.db.profile
	globalDB = RatingBuster.db.global

	options.args.profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(RatingBuster.db)
	options.args.profiles.order = 5

	local LibDualSpec = LibStub('LibDualSpec-1.0', true)

	if LibDualSpec then
		LibDualSpec:EnhanceDatabase(RatingBuster.db, "RatingBusterDB")
		LibDualSpec:EnhanceOptions(options.args.profiles, RatingBuster.db)
	end

	local always_buffed = RatingBuster.db:RegisterNamespace("AlwaysBuffed", {
		profile = {
			['*'] = false
		}
	})
	StatLogic:SetupAuraInfo(always_buffed.profile)
end

SLASH_RATINGBUSTER1, SLASH_RATINGBUSTER2 = "/ratingbuster", "/rb"
function SlashCmdList.RATINGBUSTER(input)
  if not input or input:trim() == "" then
		if not LibStub("AceConfigDialog-3.0").OpenFrames[addonNameWithVersion] then
			LibStub("AceConfigDialog-3.0"):Open(addonNameWithVersion)
		else
			LibStub("AceConfigDialog-3.0"):Close(addonNameWithVersion)
		end
  else
    LibStub("AceConfigCmd-3.0").HandleCommand(RatingBuster, "rb", addonNameWithVersion, input)
  end
end

-- OnEnable() called at PLAYER_LOGIN
function RatingBuster:OnEnable()
	addon:EnableHook(self.ProcessTooltip)
	-- Initialize playerLevel
	playerLevel = UnitLevel("player")
	-- for setting a new level
	self:RegisterEvent("PLAYER_LEVEL_UP")
	-- Events that require cache clearing
	self:RegisterEvent("CHARACTER_POINTS_CHANGED", clearCache) -- talent point changed
	self:RegisterBucketEvent("UNIT_AURA", 1) -- fire at most once every 1 second
end

function RatingBuster:OnDisable()
	addon:DisableHook()
end

-- event = PLAYER_LEVEL_UP
-- arg1 = New player level
function RatingBuster:PLAYER_LEVEL_UP(_, newlevel)
	playerLevel = newlevel
	clearCache()
end

-- event = UNIT_AURA
-- arg1 = List of UnitIDs in the AceBucket interval
function RatingBuster:UNIT_AURA(units)
	if units.player then
		clearCache()
	end
end

--------------------------
-- Process Tooltip Core --
--------------------------
--[[
"+15 Agility"
-> "+15 Agility (+0.46% Crit)"
"+15 Crit Rating"
-> "+15 Crit Rating (+1.20%)"
"Equip: Increases your hit rating by 10."
-> "Equip: Increases your hit rating by 10 (+1.20%)."
--]]
-- Empty Sockets
local EmptySocketLookup = {
	[EMPTY_SOCKET_RED] = "sumGemRed", -- EMPTY_SOCKET_RED = "Red Socket";
	[EMPTY_SOCKET_YELLOW] = "sumGemYellow", -- EMPTY_SOCKET_YELLOW = "Yellow Socket";
	[EMPTY_SOCKET_BLUE] = "sumGemBlue", -- EMPTY_SOCKET_BLUE = "Blue Socket";
	[EMPTY_SOCKET_META] = "sumGemMeta", -- EMPTY_SOCKET_META = "Meta Socket";
}

-- Avoidance Diminishing Returns
local summaryFunc = {}
local equippedSum = setmetatable({}, {
	__index = function() return 0 end
})
local equippedDodge, equippedParry, equippedMissed
local processedDodge, processedParry, processedMissed

local scanningTooltipOwners = {
	["WorldFrame"] = true,
	["UIParent"] = true,
}

function RatingBuster.ProcessTooltip(tooltip)
	-- Do nothing if the tooltip is being used as a hidden scanning tooltip
	if tooltip:GetAnchorType() == "ANCHOR_NONE" and (not tooltip:GetOwner() or scanningTooltipOwners[tooltip:GetOwner():GetName()]) then
		return
	end

	if not tooltip.GetItem then return end
	local name, link = tooltip:GetItem()
	if not name then return end

	---------------------------
	-- Set calculation level --
	---------------------------
	calcLevel = globalDB.customLevel or 0
	if calcLevel == 0 then
		calcLevel = playerLevel
	end
	if globalDB.useRequiredLevel and link then
		local _, _, _, _, reqLevel = GetItemInfo(link)
		--RatingBuster:Print(link..", "..calcLevel)
		if reqLevel and calcLevel < reqLevel then
			calcLevel = reqLevel
		end
	end
	---------------------
	-- Tooltip Scanner --
	---------------------
	-- Get equipped item avoidances
	if profileDB.enableAvoidanceDiminishingReturns then
		local red = profileDB.sumGemRed.gemID
		local yellow = profileDB.sumGemYellow.gemID
		local blue = profileDB.sumGemBlue.gemID
		local meta = profileDB.sumGemMeta.gemID
		local _, _, difflink1 = StatLogic:GetDiffID(tooltip, globalDB.sumIgnoreEnchant, globalDB.sumIgnoreGems, globalDB.sumIgnoreExtraSockets, red, yellow, blue, meta)
		StatLogic:GetSum(difflink1, equippedSum)
		equippedSum[StatLogic.Stats.Strength] = equippedSum[StatLogic.Stats.Strength] * GSM("MOD_STR")
		equippedSum[StatLogic.Stats.Agility] = equippedSum[StatLogic.Stats.Agility] * GSM("MOD_AGI")
		equippedDodge = summaryFunc[StatLogic.Stats.DodgeBeforeDR](equippedSum, "sum", difflink1) * -1
		equippedParry = summaryFunc[StatLogic.Stats.ParryBeforeDR](equippedSum, "sum", difflink1) * -1
		equippedMissed = summaryFunc[StatLogic.Stats.MissBeforeDR](equippedSum, "sum", difflink1) * -1
		processedDodge = equippedDodge
		processedParry = equippedParry
		processedMissed = equippedMissed
	else
		equippedDodge = 0
		equippedParry = 0
		equippedMissed = 0
		processedDodge = 0
		processedParry = 0
		processedMissed = 0
	end
	-- Loop through tooltip lines starting at line 2
	local tipTextLeft = tooltip:GetName().."TextLeft"
	for i = 2, tooltip:NumLines() do
		local fontString = _G[tipTextLeft..i]
		local text = fontString:GetText()
		if text then
			local color = CreateColor(fontString:GetTextColor())
			text = RatingBuster:ProcessLine(text, link, color)
			if text then
				fontString:SetText(text)
			end
		end
	end
	----------------------------
	-- Item Level and Item ID --
	----------------------------
	if (globalDB.showItemLevel or globalDB.showItemID) and link then
		if cache[link] then
			tooltip:AddLine(cache[link])
		else
			-- Get the Item ID from the link string
			local _, link, _, level = GetItemInfo(link)
			if link then
				local _, _, id = link:find("item:(%d+)")
				local newLine = ""
				local statColor = globalDB.sumStatColor
				local valueColor = globalDB.sumValueColor
				if level and globalDB.showItemLevel then
					newLine = newLine .. statColor:WrapTextInColorCode(L["ItemLevel: "])
					newLine = newLine .. valueColor:WrapTextInColorCode(level)
				end
				if id and globalDB.showItemID then
					if newLine ~= "" then
						newLine = newLine..", "
					end
					newLine = newLine .. statColor:WrapTextInColorCode(L["ItemID: "])
					newLine = newLine .. valueColor:WrapTextInColorCode(id)
				end
				if newLine ~= "" then
					cache[link] = newLine
					tooltip:AddLine(newLine)
				end
			end
		end
	end
	------------------
	-- Stat Summary --
	------------------
	--[[
	----------------
	-- Base Stats --
	----------------
	-- Health - HEALTH, STA
	-- Mana - MANA, INT
	-- Attack Power - AP, STR, AGI
	-- Ranged Attack Power - RANGED_AP, INT, AP, STR, AGI
	-- Feral Attack Power - FERAL_AP, AP, STR, AGI
	-- Spell Damage - SPELL_DMG, STA, INT, SPI
	-- Holy Damage - HOLY_SPELL_DMG, SPELL_DMG, INT, SPI
	-- Arcane Damage - ARCANE_SPELL_DMG, SPELL_DMG, INT
	-- Fire Damage - FIRE_SPELL_DMG, SPELL_DMG, STA, INT
	-- Nature Damage - NATURE_SPELL_DMG, SPELL_DMG, INT
	-- Frost Damage - FROST_SPELL_DMG, SPELL_DMG, INT
	-- Shadow Damage - SHADOW_SPELL_DMG, SPELL_DMG, STA, INT, SPI
	-- Healing - HEAL, STR, INT, SPI
	-- Hit Chance - MELEE_HIT_RATING
	-- Crit Chance - MELEE_CRIT_RATING, AGI
	-- Spell Hit Chance - SPELL_HIT_RATING
	-- Spell Crit Chance - SPELL_CRIT_RATING, INT
	-- Mana Regen - MANA_REG, SPI
	-- Health Regen - HEALTH_REG
	-- Mana Regen Not Casting - SPI
	-- Health Regen While Casting - SPI
	-- Armor - ARMOR, ARMOR_BONUS, AGI, INT
	-- Block Value - BLOCK_VALUE, STR
	-- Dodge Chance - DODGE_RATING, DEFENSE_RATING, AGI
	-- Parry Chance - PARRY_RATING, DEFENSE_RATING
	-- Block Chance - BLOCK_RATING, DEFENSE_RATING
	-- Hit Avoidance - DEFENSE_RATING, MELEE_HIT_AVOID_RATING
	-- Crit Avoidance - DEFENSE_RATING, RESILIENCE_RATING, MELEE_CRIT_AVOID_RATING
	-- Dodge Reduction - EXPERTISE_RATING
	-- Parry Reduction - EXPERTISE_RATING
	---------------------
	-- Composite Stats --
	---------------------
	-- Strength - STR
	-- Agility - AGI
	-- Stamina - STA
	-- Intellect - INT
	-- Spirit - SPI
	-- Defense - DEFENSE_RATING
	-- Weapon Skill
	-- Expertise - EXPERTISE_RATING
	--]]
	if globalDB.showSum then
		RatingBuster:StatSummary(tooltip, link)
	end
	---------------------
	-- Repaint tooltip --
	---------------------
	tooltip:Show()
end

function RatingBuster:ProcessLine(text, link, color)
	-- Get data from cache if available
	local cacheID = text..calcLevel
	local cacheText = cache[cacheID]
	if cacheText then
		if cacheText ~= text then
			return cacheText
		end
	elseif EmptySocketLookup[text] and profileDB[EmptySocketLookup[text]].gemText then -- Replace empty sockets with gem text
		text = profileDB[EmptySocketLookup[text]].gemText
		cache[cacheID] = text
		-- SetText
		return text
	elseif text:find("%d") then -- do nothing if we don't find a number
		-- Temporarily replace exclusions
		local exclusions = false
		for exclusion, replacement in pairs(L["exclusions"]) do
			local count
			text, count = text:gsub(exclusion, replacement)
			if count > 0 then
				exclusions = true
			end
		end
		-- Initial pattern check, do nothing if not found
		-- Check for separators and bulid separatorTable
		local separatorTable = {}
		for _, sep in ipairs(L["separators"]) do
			if text:find(sep) then
				tinsert(separatorTable, sep)
			end
		end
		-- RecursivelySplitLine
		text = RatingBuster:RecursivelySplitLine(text, separatorTable, link, color)
		-- Revert exclusions
		if exclusions then
			for exclusion, replacement in pairs(L["exclusions"]) do
				text = text:gsub(replacement, exclusion)
			end
		end
		cache[cacheID] = text
		-- SetText
		return text
	end
end

---------------------------------------------------------------------------------
-- Recursive algorithm that divides a string into pieces using the separators in separatorTable,
-- processes them separately, then joins them back together
---------------------------------------------------------------------------------
-- text = "+24 Agility/+4 Stamina and +4 Spell Crit/+5 Spirit"
-- separatorTable = {"/", " and ", ","}
-- RatingBuster:RecursivelySplitLine("+24 Agility/+4 Stamina, +4 Dodge and +4 Spell Crit/+5 Spirit", {"/", " and ", ",", "%. ", " for ", "&"})
-- RatingBuster:RecursivelySplitLine("+6法術傷害及5耐力", {"/", "和", ",", "。", " 持續 ", "&", "及",})
function RatingBuster:RecursivelySplitLine(text, separatorTable, link, color)
	if type(separatorTable) == "table" and table.maxn(separatorTable) > 0 then
		local sep = tremove(separatorTable, 1)
		text =  text:gsub(sep, "@")
		text = strsplittable("@", text)
		local processedText = {}
		local tempTable = {}
		for _, t in ipairs(text) do
			copyTable(tempTable, separatorTable)
			tinsert(processedText, self:RecursivelySplitLine(t, tempTable, link, color))
		end
		-- Join text
		return (table.concat(processedText, "@"):gsub("@", sep))
	else
		return self:ProcessText(text, link, color)
	end
end

function RatingBuster:ProcessText(text, link, color)
	-- Convert text to lower so we don't have to worry about same ratings with different cases
	local lowerText = text:lower()
	-- Check if text has a matching pattern
	for _, num in ipairs(L["numberPatterns"]) do
		-- Capture the stat value
		local _, insertionPoint, value, partialtext = lowerText:find(num.pattern)
		if value then
			-- Check and switch captures if needed
			if partialtext and tonumber(partialtext) then
				value, partialtext = partialtext, value
			end
			-- Capture the stat name
			for _, statPattern in ipairs(L["statList"]) do
				local pattern, stat = unpack(statPattern)
				if (not partialtext and lowerText:find(pattern)) or (partialtext and partialtext:find(pattern)) then
					value = tonumber(value)
					local infoString = RatingBuster:ProcessStat(stat, value, link, color)
					if infoString ~= "" then
						-- Change insertion point if necessary
						if num.addInfo == "AfterStat" then
							local _, statInsertionPoint = lowerText:find(pattern)
							if statInsertionPoint > insertionPoint then
								insertionPoint = statInsertionPoint
							end
						end

						-- Backwards Compatibility
						if not globalDB.textColor.GenerateHexColorMarkup then
							local old = globalDB.textColor
							if type(old) == "table" and old.r and old.g and old.b then
								globalDB.textColor = CreateColor(old.r, old.g, old.b)
							else
								globalDB.textColor = defaults.global.textColor
							end
						end

						-- Insert info into text. table.concat should be more efficient than many .. concats
						return table.concat({
							text:sub(1, insertionPoint),
							" ",
							globalDB.textColor:GenerateHexColorMarkup(),
							"(",
							infoString,
							")",
							"|r",
							text:sub(insertionPoint + 1)
						})
					else
						return text
					end
				end
			end
		end
	end
	return text
end

do
	local RatingType = {
		Melee = {
			[StatLogic.Stats.MeleeHitRating] = true,
			[StatLogic.Stats.MeleeCritRating] = true,
			[StatLogic.Stats.MeleeHasteRating] = true,
		},
		Ranged = {
			[StatLogic.Stats.RangedHitRating] = true,
			[StatLogic.Stats.RangedCritRating] = true,
			[StatLogic.Stats.RangedHasteRating] = true,
		},
		Spell = {
			[StatLogic.Stats.SpellHitRating] = true,
			[StatLogic.Stats.SpellCritRating] = true,
			[StatLogic.Stats.SpellHasteRating] = true,
		},
		Decimal = {
			[StatLogic.Stats.DefenseRating] = true,
			[StatLogic.Stats.ExpertiseRating] = true,
		}
	}

	function RatingBuster:ProcessStat(statID, value, link, color)
		local infoString = ""
		if StatLogic.GenericStatMap[statID] then
			local statList = StatLogic.GenericStatMap[statID]
			local first = true
			for _, convertedStatID in ipairs(statList) do
				if not RatingType.Ranged[convertedStatID] then
					local result = RatingBuster:ProcessStat(convertedStatID, value)
					if result and result ~= "" then
						if not first then
							infoString = infoString .. ", "
						end
						infoString = infoString .. result
						first = false
					end
				end
			end
		elseif StatLogic.RatingBase[statID] and profileDB.showRatings then
			--------------------
			-- Combat Ratings --
			--------------------
			-- Calculate stat value
			local effect = StatLogic:GetEffectFromRating(value, statID, calcLevel)
			--self:Debug(reversedAmount..", "..amount..", "..v[2]..", "..RatingBuster.targetLevel)-- debug
			-- If rating is resilience, add a minus sign
			if statID == StatLogic.Stats.DefenseRating and profileDB.defBreakDown then
				effect = effect * 0.04
				processedDodge = processedDodge + effect
				processedMissed = processedMissed + effect
				local numStats = 5
				if GetParryChance() == 0 then
					numStats = numStats - 1
				else
					processedParry = processedParry + effect
				end
				if GetBlockChance() == 0 then
					numStats = numStats - 1
				end
				infoString = ("%+.2f%% x"..numStats):format(effect)
			elseif statID == StatLogic.Stats.DodgeRating and profileDB.enableAvoidanceDiminishingReturns then
				infoString = ("%+.2f%%"):format(StatLogic:GetAvoidanceGainAfterDR(StatLogic.Stats.Dodge, processedDodge + effect) - StatLogic:GetAvoidanceGainAfterDR(StatLogic.Stats.Dodge, processedDodge))
				processedDodge = processedDodge + effect
			elseif statID == StatLogic.Stats.ParryRating and profileDB.enableAvoidanceDiminishingReturns then
				infoString = ("%+.2f%%"):format(StatLogic:GetAvoidanceGainAfterDR(StatLogic.Stats.Parry, processedParry + effect) - StatLogic:GetAvoidanceGainAfterDR(StatLogic.Stats.Parry, processedParry))
				processedParry = processedParry + effect
			elseif statID == StatLogic.Stats.ExpertiseRating and profileDB.expBreakDown then
				if addon.tocversion < 30000 then
					-- Expertise is truncated in TBC but not in Wrath
					effect = floor(effect)
				end
				effect = effect * -0.25
				if profileDB.detailedConversionText then
					infoString = L["$value to be Dodged/Parried"]:gsub("$value", ("%+.2f%%%%"):format(effect))
				else
					infoString = ("%+.2f%%"):format(effect)
				end
			elseif statID == StatLogic.Stats.ResilienceRating then -- Resilience
				effect = effect * -1
				if profileDB.detailedConversionText then
					local infoTable = {}

					if addon.tocversion >= 30000 then
						-- Wrath
						tinsert(infoTable, (L["$value to be Crit"]:gsub("$value", ("%+.2f%%%%"):format(effect))))
						tinsert(infoTable, (L["$value Crit Dmg Taken"]:gsub("$value", ("%+.2f%%%%"):format(effect * RESILIENCE_CRIT_CHANCE_TO_DAMAGE_REDUCTION_MULTIPLIER))))
						tinsert(infoTable, (L["$value Dmg Taken"]:gsub("$value", ("%+.2f%%%%"):format(effect * RESILIENCE_CRIT_CHANCE_TO_CONSTANT_DAMAGE_REDUCTION_MULTIPLIER))))
					elseif addon.tocversion >= 20000 then
						-- TBC
						tinsert(infoTable, (L["$value to be Crit"]:gsub("$value", ("%+.2f%%%%"):format(effect))))
						tinsert(infoTable, (L["$value Crit Dmg Taken"]:gsub("$value", ("%+.2f%%%%"):format(effect * 2))))
						tinsert(infoTable, (L["$value DOT Dmg Taken"]:gsub("$value", ("%+.2f%%%%"):format(effect))))
					end

					infoString = table.concat(infoTable, ", ")
				else
					infoString = ("%+.2f%%"):format(effect)
				end
			else
				local pattern = "%+.2f%%"
				local show = false
				if RatingType.Melee[statID] then
					if profileDB.ratingPhysical then
						show = true
					end
				elseif RatingType.Spell[statID] then
					if profileDB.ratingSpell then
						if not profileDB.ratingPhysical then
							show = true
						elseif ( statID == StatLogic.Stats.SpellHitRating or (statID == StatLogic.Stats.SpellHasteRating and StatLogic.ExtraHasteClasses[class])) then
							show = true
							pattern = L["$value Spell"]:gsub("$value", "%%+.2f%%%%")
						end
					end
				elseif RatingType.Decimal[statID] then
					show = true
					pattern = "%+.2f"
				else
					show = true
				end

				if show then
					infoString = pattern:format(effect)
				end
			end
		elseif statID == StatLogic.Stats.Strength and profileDB.showStats then
			--------------
			-- Strength --
			--------------
			local statmod = 1
			if profileDB.enableStatMods then
				statmod = GSM("MOD_STR")
				value = value * statmod
			end
			local infoTable = {}
			if profileDB.showAPFromStr then
				local mod = GSM("MOD_AP")
				local effect = value * GSM("ADD_AP_MOD_STR") * mod
				if (mod ~= 1 or statmod ~= 1) and floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value AP"]:gsub("$value", ("%+.1f"):format(effect))))
				elseif floor(abs(effect) + 0.5) > 0 then -- so we don't get +0 AP when effect < 0.5
					tinsert(infoTable, (L["$value AP"]:gsub("$value", ("%+.0f"):format(effect))))
				end
			end
			if profileDB.showBlockValueFromStr then
				local effect = value * GSM("ADD_BLOCK_VALUE_MOD_STR")
				if floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value Block"]:gsub("$value", ("%+.1f"):format(effect))))
				end
			end
			-- Shaman: Mental Quickness
			-- Paladin: Sheath of Light, Touched by the Light
			if profileDB.showSpellDmgFromStr then
				local mod = GSM("MOD_AP") * GSM("MOD_SPELL_DMG")
				local effect = (value * GSM("ADD_AP_MOD_STR") * GSM("ADD_SPELL_DMG_MOD_AP")
					+ value * GSM("ADD_SPELL_DMG_MOD_STR")) * mod
				if (mod ~= 1 or statmod ~= 1) and floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value Spell Dmg"]:gsub("$value", ("%+.1f"):format(effect))))
				elseif floor(abs(effect) + 0.5) > 0 then
					tinsert(infoTable, (L["$value Spell Dmg"]:gsub("$value", ("%+.0f"):format(effect))))
				end
			end
			if profileDB.showHealingFromStr then
				local mod = GSM("MOD_AP") * GSM("MOD_HEALING")
				local effect = (value * GSM("ADD_AP_MOD_STR") * GSM("ADD_HEALING_MOD_AP")
					+ value * GSM("ADD_HEALING_MOD_STR")) * mod
				if (mod ~= 1 or statmod ~= 1) and floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value Heal"]:gsub("$value", ("%+.1f"):format(effect))))
				elseif floor(abs(effect) + 0.5) > 0 then
					tinsert(infoTable, (L["$value Heal"]:gsub("$value", ("%+.0f"):format(effect))))
				end
			end
			-- Death Knight: Forceful Deflection - Passive
			if profileDB.showParryFromStr then
				local rating = value * GSM("ADD_PARRY_RATING_MOD_STR")
				local effect = StatLogic:GetEffectFromRating(rating, StatLogic.Stats.ParryRating, calcLevel)
				if profileDB.enableAvoidanceDiminishingReturns then
					local effectNoDR = effect
					effect = StatLogic:GetAvoidanceGainAfterDR(StatLogic.Stats.Parry, processedParry + effect) - StatLogic:GetAvoidanceGainAfterDR(StatLogic.Stats.Parry, processedParry)
					processedParry = processedParry + effectNoDR
				end
				if effect > 0 then
					tinsert(infoTable, (L["$value% Parry"]:gsub("$value", ("%+.2f"):format(effect))))
				end
			else
				local rating = value * GSM("ADD_PARRY_RATING_MOD_STR")
				local effect = StatLogic:GetEffectFromRating(rating, StatLogic.Stats.ParryRating, calcLevel)
				processedParry = processedParry + effect
			end
			infoString = table.concat(infoTable, ", ")
		elseif statID == StatLogic.Stats.Agility and profileDB.showStats then
			-------------
			-- Agility --
			-------------
			local statmod = 1
			if profileDB.enableStatMods then
				statmod = GSM("MOD_AGI")
				value = value * statmod
			end
			local infoTable = {}
			if profileDB.showAPFromAgi then
				local mod = GSM("MOD_AP")
				local effect = value * GSM("ADD_AP_MOD_AGI") * mod
				if (mod ~= 1 or statmod ~= 1) and floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value AP"]:gsub("$value", ("%+.1f"):format(effect))))
				elseif floor(abs(effect) + 0.5) > 0 then
					tinsert(infoTable, (L["$value AP"]:gsub("$value", ("%+.0f"):format(effect))))
				end
			end
			if profileDB.showRAPFromAgi then
				local mod = GSM("MOD_RANGED_AP")
				local effect = value * GSM("ADD_RANGED_AP_MOD_AGI") * mod
				if (mod ~= 1 or statmod ~= 1) and floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value RAP"]:gsub("$value", ("%+.1f"):format(effect))))
				elseif floor(abs(effect) + 0.5) > 0 then
					tinsert(infoTable, (L["$value RAP"]:gsub("$value", ("%+.0f"):format(effect))))
				end
			end
			if profileDB.showCritFromAgi then
				local effect = value * StatLogic:GetCritPerAgi(class, calcLevel)
				if effect > 0 then
					tinsert(infoTable, (L["$value% Crit"]:gsub("$value", ("%+.2f"):format(effect))))
				end
			end
			if profileDB.showDodgeFromAgi and (calcLevel == playerLevel) then
				local effect = value * StatLogic:GetDodgePerAgi()
				if effect > 0 then
					tinsert(infoTable, (L["$value% Dodge"]:gsub("$value", ("%+.2f"):format(effect))))
				end
			end
			if profileDB.showArmorFromAgi then
				local effect = value * 2
				if effect > 0 then
					tinsert(infoTable, (L["$value Armor"]:gsub("$value", ("%+.0f"):format(effect))))
				end
			end
			-- Shaman: Mental Quickness
			-- Paladin: Sheath of Light, Touched by the Light
			if profileDB.showSpellDmgFromAgi then
				local mod = GSM("MOD_AP") * GSM("MOD_SPELL_DMG")
				local effect = (value * GSM("ADD_AP_MOD_AGI") * GSM("ADD_SPELL_DMG_MOD_AP")) * mod
				if (mod ~= 1 or statmod ~= 1) and floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value Spell Dmg"]:gsub("$value", ("%+.1f"):format(effect))))
				elseif floor(abs(effect) + 0.5) > 0 then
					tinsert(infoTable, (L["$value Spell Dmg"]:gsub("$value", ("%+.0f"):format(effect))))
				end
			end
			-- Druid: Nurturing Instinct
			if profileDB.showHealingFromAgi then
				local mod = GSM("MOD_AP") * GSM("MOD_HEALING")
				local effect = (value * GSM("ADD_AP_MOD_AGI") * GSM("ADD_HEALING_MOD_AP")
					+ value * GSM("ADD_HEALING_MOD_AGI") / GSM("MOD_AP")) * mod
				if (mod ~= 1 or statmod ~= 1) and floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value Heal"]:gsub("$value", ("%+.1f"):format(effect))))
				elseif floor(abs(effect) + 0.5) > 0 then
					tinsert(infoTable, (L["$value Heal"]:gsub("$value", ("%+.0f"):format(effect))))
				end
			end
			infoString = table.concat(infoTable, ", ")
		elseif statID == StatLogic.Stats.Stamina and profileDB.showStats then
			-------------
			-- Stamina --
			-------------
			local statmod = 1
			if profileDB.enableStatMods then
				statmod = GSM("MOD_STA")
				value = value * statmod
			end
			local infoTable = {}
			if profileDB.showHealthFromSta then
				local mod = GSM("MOD_HEALTH")
				local effect = value * GSM("ADD_HEALTH_MOD_STA") * mod
				if (mod ~= 1 or statmod ~= 1) and floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value HP"]:gsub("$value", ("%+.1f"):format(effect))))
				elseif floor(abs(effect) + 0.5) > 0 then
					tinsert(infoTable, (L["$value HP"]:gsub("$value", ("%+.0f"):format(effect))))
				end
			end
			if profileDB.showSpellDmgFromSta then
				local mod = GSM("MOD_SPELL_DMG")
				local effect = value * mod * (GSM("ADD_SPELL_DMG_MOD_STA")
					+ GSM("ADD_SPELL_DMG_MOD_PET_STA") * GSM("MOD_PET_STA") * GSM("ADD_PET_STA_MOD_STA"))
				if floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value Spell Dmg"]:gsub("$value", ("%+.1f"):format(effect))))
				end
			end
			-- "ADD_AP_MOD_STA" -- Hunter: Hunter vs. Wild
			if profileDB.showAPFromSta then
				local mod = GSM("MOD_AP")
				local effect = value * GSM("ADD_AP_MOD_STA") * mod
				if (mod ~= 1 or statmod ~= 1) and floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value AP"]:gsub("$value", ("%+.1f"):format(effect))))
				elseif floor(abs(effect) + 0.5) > 0 then
					tinsert(infoTable, (L["$value AP"]:gsub("$value", ("%+.0f"):format(effect))))
				end
			end
			infoString = table.concat(infoTable, ", ")
		elseif statID == StatLogic.Stats.Intellect and profileDB.showStats then
			---------------
			-- Intellect --
			---------------
			local statmod = 1
			if profileDB.enableStatMods then
				statmod = GSM("MOD_INT")
				value = value * statmod
			end
			local infoTable = {}
			if profileDB.showManaFromInt then
				local mod = GSM("MOD_MANA")
				local effect = value * GSM("ADD_MANA_MOD_INT") * mod -- 15 Mana per Int
				if (mod ~= 1 or statmod ~= 1) and floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value MP"]:gsub("$value", ("%+.1f"):format(effect))))
				elseif floor(abs(effect) + 0.5) > 0 then
					tinsert(infoTable, (L["$value MP"]:gsub("$value", ("%+.0f"):format(effect))))
				end
			end
			if profileDB.showSpellCritFromInt then
				local effect = value * StatLogic:GetSpellCritPerInt(class, calcLevel)
				if effect > 0 then
					tinsert(infoTable, (L["$value% Spell Crit"]:gsub("$value", ("%+.2f"):format(effect))))
				end
			end
			if profileDB.showSpellDmgFromInt then
				local mod = GSM("MOD_SPELL_DMG")
				local effect = value * mod * (GSM("ADD_SPELL_DMG_MOD_INT")
					+ GSM("ADD_SPELL_DMG_MOD_PET_INT") * GSM("MOD_PET_INT") * GSM("ADD_PET_INT_MOD_INT"))
				if floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value Spell Dmg"]:gsub("$value", ("%+.1f"):format(effect))))
				end
			end
			if profileDB.showHealingFromInt then
				local mod = GSM("MOD_HEALING")
				local effect = value * GSM("ADD_HEALING_MOD_INT") * mod
				if floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value Heal"]:gsub("$value", ("%+.1f"):format(effect))))
				end
			end
			if profileDB.showMP5FromInt then
				local effect = value * GSM("ADD_MANA_REG_MOD_INT")
					+ value * GSM("ADD_NORMAL_MANA_REG_MOD_INT") * GSM("MOD_NORMAL_MANA_REG") * math.min(GSM("ADD_MANA_REG_MOD_NORMAL_MANA_REG"), 1)
					+ value * GSM("ADD_MANA_MOD_INT") * GSM("MOD_MANA") * GSM("ADD_MANA_REG_MOD_MANA") -- Replenishment
				if floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value MP5"]:gsub("$value", ("%+.1f"):format(effect))))
				end
			end
			if profileDB.showMP5NCFromInt then
				local effect = value * GSM("ADD_MANA_REG_MOD_INT")
					+ value * GSM("ADD_NORMAL_MANA_REG_MOD_INT") * GSM("MOD_NORMAL_MANA_REG")
					+ value * GSM("ADD_MANA_MOD_INT") * GSM("MOD_MANA") * GSM("ADD_MANA_REG_MOD_MANA") -- Replenishment
				if floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value MP5(NC)"]:gsub("$value", ("%+.1f"):format(effect))))
				end
			end
			if profileDB.showRAPFromInt then
				local mod = GSM("MOD_RANGED_AP")
				local effect = value * GSM("ADD_RANGED_AP_MOD_INT") * mod
				if floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value RAP"]:gsub("$value", ("%+.1f"):format(effect))))
				end
			end
			if profileDB.showArmorFromInt then
				local effect = value * GSM("ADD_BONUS_ARMOR_MOD_INT")
				if floor(abs(effect) + 0.5) > 0 then
					tinsert(infoTable, (L["$value Armor"]:gsub("$value", ("%+.0f"):format(effect))))
				end
			end
			-- "ADD_AP_MOD_INT" -- Shaman: Mental Dexterity
			if profileDB.showAPFromInt then
				local mod = GSM("MOD_AP")
				local effect = value * GSM("ADD_AP_MOD_INT") * mod
				if (mod ~= 1 or statmod ~= 1) and floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value AP"]:gsub("$value", ("%+.1f"):format(effect))))
				elseif floor(abs(effect) + 0.5) > 0 then
					tinsert(infoTable, (L["$value AP"]:gsub("$value", ("%+.0f"):format(effect))))
				end
			end
			infoString = table.concat(infoTable, ", ")
		elseif statID == StatLogic.Stats.Spirit and profileDB.showStats then
			------------
			-- Spirit --
			------------
			local statmod = 1
			if profileDB.enableStatMods then
				statmod = GSM("MOD_SPI")
				value = value * statmod
			end
			local infoTable = {}
			if profileDB.showMP5FromSpi then
				local effect = value * GSM("ADD_NORMAL_MANA_REG_MOD_SPI") * GSM("MOD_NORMAL_MANA_REG") * math.min(GSM("ADD_MANA_REG_MOD_NORMAL_MANA_REG"), 1)
				if floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value MP5"]:gsub("$value", ("%+.1f"):format(effect))))
				end
			end
			if profileDB.showMP5NCFromSpi then
				local effect = value * GSM("ADD_NORMAL_MANA_REG_MOD_SPI") * GSM("MOD_NORMAL_MANA_REG")
				if floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value MP5(NC)"]:gsub("$value", ("%+.1f"):format(effect))))
				end
			end
			if profileDB.showHP5FromSpi then
				local effect = value * GSM("ADD_NORMAL_HEALTH_REG_MOD_SPI") * GSM("MOD_NORMAL_HEALTH_REG") * GSM("ADD_HEALTH_REG_MOD_NORMAL_HEALTH_REG")
				if floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value HP5"]:gsub("$value", ("%+.1f"):format(effect))))
				end
			end
			if profileDB.showHP5NCFromSpi then
				local effect = value * GSM("ADD_NORMAL_HEALTH_REG_MOD_SPI") * GSM("MOD_NORMAL_HEALTH_REG")
				if floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value HP5(NC)"]:gsub("$value", ("%+.1f"):format(effect))))
				end
			end
			if profileDB.showSpellDmgFromSpi then
				local mod = GSM("MOD_SPELL_DMG")
				local effect = value * GSM("ADD_SPELL_DMG_MOD_SPI") * mod
				if floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value Spell Dmg"]:gsub("$value", ("%+.1f"):format(effect))))
				end
			end
			if profileDB.showHealingFromSpi then
				local mod = GSM("MOD_HEALING")
				local effect = value * GSM("ADD_HEALING_MOD_SPI") * mod
				if floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value Heal"]:gsub("$value", ("%+.1f"):format(effect))))
				end
			end
			if profileDB.showSpellCritFromSpi then
				local mod = GSM("ADD_SPELL_CRIT_RATING_MOD_SPI")
				local effect = StatLogic:GetEffectFromRating(value * mod, StatLogic.Stats.SpellCritRating, calcLevel)
				if effect > 0 then
					tinsert(infoTable, (L["$value% Spell Crit"]:gsub("$value", ("%+.2f"):format(effect))))
				end
			end
			infoString = table.concat(infoTable, ", ")
		elseif profileDB.showAPFromArmor and statID == StatLogic.Stats.Armor then
			-----------
			-- Armor --
			-----------
			if profileDB.enableStatMods then
				local base, bonus = StatLogic:GetArmorDistribution(link, value, color)
				value = base * GSM("MOD_ARMOR") + bonus
			end
			local infoTable = {}
			local effect = value * GSM("ADD_AP_MOD_ARMOR") * GSM("MOD_AP")
			if floor(abs(effect) * 10 + 0.5) > 0 then
				tinsert(infoTable, (L["$value AP"]:gsub("$value", ("%+.1f"):format(effect))))
			end
			infoString = table.concat(infoTable, ", ")
		elseif statID == StatLogic.Stats.AttackPower then
			------------------
			-- Attack Power --
			------------------
			local statmod = 1
			if profileDB.enableStatMods then
				statmod = GSM("MOD_AP")
				value = value * statmod
			end
			local infoTable = {}
			-- Shaman: Mental Quickness
			-- Paladin: Sheath of Light, Touched by the Light
			if profileDB.showSpellDmgFromAP then
				local mod = GSM("MOD_AP") * GSM("MOD_SPELL_DMG")
				local effect = (value * GSM("ADD_SPELL_DMG_MOD_AP")) * mod
				if (mod ~= 1 or statmod ~= 1) and floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value Spell Dmg"]:gsub("$value", ("%+.1f"):format(effect))))
				elseif floor(abs(effect) + 0.5) > 0 then
					tinsert(infoTable, (L["$value Spell Dmg"]:gsub("$value", ("%+.0f"):format(effect))))
				end
			end
			if profileDB.showHealingFromAP then
				local mod = GSM("MOD_AP") * GSM("MOD_HEALING")
				local effect = (value * GSM("ADD_HEALING_MOD_AP")) * mod
				if (mod ~= 1 or statmod ~= 1) and floor(abs(effect) * 10 + 0.5) > 0 then
					tinsert(infoTable, (L["$value Heal"]:gsub("$value", ("%+.1f"):format(effect))))
				elseif floor(abs(effect) + 0.5) > 0 then
					tinsert(infoTable, (L["$value Heal"]:gsub("$value", ("%+.0f"):format(effect))))
				end
			end
			infoString = table.concat(infoTable, ", ")
		end
		return infoString
	end
end

local blizzardComparisonPatterns = {
	[ITEM_DELTA_DESCRIPTION] = true,
	[ITEM_DELTA_MULTIPLE_COMPARISON_DESCRIPTION] = true,
}

local function RemoveFontString(fontString)
	fontString:SetText("")
	fontString:Hide()
	fontString:SetWidth(0)
end

local function RemoveBlizzardItemComparisons(tooltip)
	if not tooltip or not globalDB.hideBlizzardComparisons then return end

	for _, shoppingTooltip in pairs(tooltip.shoppingTooltips) do
		local tipTextLeft = shoppingTooltip:GetName().."TextLeft"
		local tipTextRight = shoppingTooltip:GetName().."TextRight"
		local isBlizzardComparison = false

		for i = 2, shoppingTooltip:NumLines() do
			local fontString = _G[tipTextLeft..i]
			local text = fontString:GetText()
			if text then
				if not isBlizzardComparison and blizzardComparisonPatterns[text] then
					isBlizzardComparison = true
					local previousFontString = _G[tipTextLeft .. (i-1)]
					if previousFontString:GetText():find("^%s*") then
						RemoveFontString(previousFontString)
					end
				end

				if isBlizzardComparison then
					RemoveFontString(fontString)
					RemoveFontString(_G[tipTextRight..i])
				end
			end
		end
		shoppingTooltip:Show()
	end
end

hooksecurefunc("GameTooltip_ShowCompareItem", RemoveBlizzardItemComparisons)

-- Color Numbers
local GREEN_FONT_COLOR_CODE = GREEN_FONT_COLOR_CODE -- "|cff20ff20" Green
local HIGHLIGHT_FONT_COLOR_CODE = HIGHLIGHT_FONT_COLOR_CODE -- "|cffffffff" White
local RED_FONT_COLOR_CODE = RED_FONT_COLOR_CODE -- "|cffffffff" White
local FONT_COLOR_CODE_CLOSE = FONT_COLOR_CODE_CLOSE -- "|r"
local function colorNum(text, num)
	if num > 0 then
		return GREEN_FONT_COLOR_CODE..text..FONT_COLOR_CODE_CLOSE
	elseif num < 0 then
		return RED_FONT_COLOR_CODE..text..FONT_COLOR_CODE_CLOSE
	else
		return HIGHLIGHT_FONT_COLOR_CODE..text..FONT_COLOR_CODE_CLOSE
	end
end

-- Used armor type each class uses
local classArmorTypes = {
	WARRIOR = {
		[Enum.ItemArmorSubclass["Plate"]] = true,
		[Enum.ItemArmorSubclass["Mail"]] = true,
		[Enum.ItemArmorSubclass["Leather"]] = true,
	},
	PALADIN = {
		[Enum.ItemArmorSubclass["Plate"]] = true,
		[Enum.ItemArmorSubclass["Mail"]] = true,
		[Enum.ItemArmorSubclass["Leather"]] = true,
		[Enum.ItemArmorSubclass["Cloth"]] = true,
	},
	HUNTER = {
		[Enum.ItemArmorSubclass["Mail"]] = true,
		[Enum.ItemArmorSubclass["Leather"]] = true,
	},
	ROGUE = {
		[Enum.ItemArmorSubclass["Leather"]] = true,
	},
	PRIEST = {
		[Enum.ItemArmorSubclass["Cloth"]] = true,
	},
	DEATHKNIGHT = {
		[Enum.ItemArmorSubclass["Plate"]] = true,
		[Enum.ItemArmorSubclass["Mail"]] = true,
		[Enum.ItemArmorSubclass["Leather"]] = true,
	},
	SHAMAN = {
		[Enum.ItemArmorSubclass["Mail"]] = true,
		[Enum.ItemArmorSubclass["Leather"]] = true,
		[Enum.ItemArmorSubclass["Cloth"]] = true,
	},
	MAGE = {
		[Enum.ItemArmorSubclass["Cloth"]] = true,
	},
	WARLOCK = {
		[Enum.ItemArmorSubclass["Cloth"]] = true,
	},
	DRUID = {
		[Enum.ItemArmorSubclass["Leather"]] = true,
		[Enum.ItemArmorSubclass["Cloth"]] = true,
	},
}

local armorTypes = {
	[Enum.ItemArmorSubclass["Plate"]] = true,
	[Enum.ItemArmorSubclass["Mail"]] = true,
	[Enum.ItemArmorSubclass["Leather"]] = true,
	[Enum.ItemArmorSubclass["Cloth"]] = true,
}

-- Interface_<Expansion>/FrameXML/PaperDollFrame.lua Compatibility
if not DODGE_PARRY_BLOCK_PERCENT_PER_DEFENSE then DODGE_PARRY_BLOCK_PERCENT_PER_DEFENSE = 0.04 end

local summaryCalcData = {
	-----------
	-- Basic --
	-----------
	-- Strength - STR
	{
		option = "sumStr",
		name = StatLogic.Stats.Strength,
		func = function(sum)
			return sum[StatLogic.Stats.Strength]
		end,
	},
	-- Agility - AGI
	{
		option = "sumAgi",
		name = StatLogic.Stats.Agility,
		func = function(sum)
			return sum[StatLogic.Stats.Agility]
		end,
	},
	-- Stamina - STA
	{
		option = "sumSta",
		name = StatLogic.Stats.Stamina,
		func = function(sum)
			return sum[StatLogic.Stats.Stamina]
		end,
	},
	-- Intellect - INT
	{
		option = "sumInt",
		name = StatLogic.Stats.Intellect,
		func = function(sum)
			return sum[StatLogic.Stats.Intellect]
		end,
	},
	-- Spirit - SPI
	{
		option = "sumSpi",
		name = StatLogic.Stats.Spirit,
		func = function(sum)
			return sum[StatLogic.Stats.Spirit]
		end,
	},
	-- Health - HEALTH, STA
	{
		option = "sumHP",
		name = StatLogic.Stats.Health,
		func = function(sum)
			return (sum[StatLogic.Stats.Health] + (sum[StatLogic.Stats.Stamina] * GSM("ADD_HEALTH_MOD_STA"))) * GSM("MOD_HEALTH")
		end,
	},
	-- Mana - MANA, INT
	{
		option = "sumMP",
		name = StatLogic.Stats.Mana,
		func = function(sum)
			return (sum[StatLogic.Stats.Mana] + (sum[StatLogic.Stats.Intellect] * GSM("ADD_MANA_MOD_INT"))) * GSM("MOD_MANA")
		end,
	},
	-- Health Regen - HEALTH_REG
	{
		option = "sumHP5",
		name = StatLogic.Stats.HealthRegen,
		func = function(sum)
			return sum[StatLogic.Stats.HealthRegen]
				+ sum[StatLogic.Stats.Spirit] * GSM("ADD_NORMAL_HEALTH_REG_MOD_SPI") * GSM("MOD_NORMAL_HEALTH_REG") * GSM("ADD_HEALTH_REG_MOD_NORMAL_HEALTH_REG")
		end,
	},
	-- Health Regen while Out of Combat - HEALTH_REG, SPI
	{
		option = "sumHP5OC",
		name = StatLogic.Stats.HealthRegenOutOfCombat,
		func = function(sum)
			return sum[StatLogic.Stats.HealthRegen]
				+ sum[StatLogic.Stats.Spirit] * GSM("ADD_NORMAL_HEALTH_REG_MOD_SPI") * GSM("MOD_NORMAL_HEALTH_REG")
		end,
	},
	-- Mana Regen - MANA_REG, SPI, INT
	{
		option = "sumMP5",
		name = StatLogic.Stats.ManaRegen,
		func = function(sum)
			return sum[StatLogic.Stats.ManaRegen]
				+ sum[StatLogic.Stats.Intellect] * GSM("ADD_MANA_REG_MOD_INT")
				+ math.min(GSM("ADD_MANA_REG_MOD_NORMAL_MANA_REG"), 1) * GSM("MOD_NORMAL_MANA_REG") * (
					sum[StatLogic.Stats.Intellect] * GSM("ADD_NORMAL_MANA_REG_MOD_INT")
					+ sum[StatLogic.Stats.Spirit] * GSM("ADD_NORMAL_MANA_REG_MOD_SPI")
				) + summaryFunc[StatLogic.Stats.Mana](sum) * GSM("ADD_MANA_REG_MOD_MANA")
		end,
	},
	-- Mana Regen while Not casting - MANA_REG, SPI, INT
	{
		option = "sumMP5NC",
		name = StatLogic.Stats.ManaRegenNotCasting,
		func = function(sum)
			return sum[StatLogic.Stats.ManaRegen]
				+ sum[StatLogic.Stats.Intellect] * GSM("ADD_MANA_REG_MOD_INT")
				+ GSM("MOD_NORMAL_MANA_REG") * (
					sum[StatLogic.Stats.Intellect] * GSM("ADD_NORMAL_MANA_REG_MOD_INT")
					+ sum[StatLogic.Stats.Spirit] * GSM("ADD_NORMAL_MANA_REG_MOD_SPI")
				) + summaryFunc[StatLogic.Stats.Mana](sum) * GSM("ADD_MANA_REG_MOD_MANA")
		end,
	},
	---------------------
	-- Physical Damage --
	---------------------
	-- Attack Power - AP, STR, AGI
	{
		option = "sumAP",
		name = StatLogic.Stats.AttackPower,
		func = function(sum)
			return GSM("MOD_AP") * (
				-- Feral Druid Predatory Strikes
				(sum[StatLogic.Stats.FeralAttackPower] > 0 and GSM("MOD_FERAL_AP") or 1) * (
					sum[StatLogic.Stats.AttackPower]
					+ sum[StatLogic.Stats.FeralAttackPower] * GSM("ADD_AP_MOD_FERAL_AP")
				) + sum[StatLogic.Stats.Strength] * GSM("ADD_AP_MOD_STR")
				+ sum[StatLogic.Stats.Agility] * GSM("ADD_AP_MOD_AGI")
				+ sum[StatLogic.Stats.Stamina] * GSM("ADD_AP_MOD_STA")
				+ sum[StatLogic.Stats.Intellect] * GSM("ADD_AP_MOD_INT")
				+ summaryFunc[StatLogic.Stats.Armor](sum) * GSM("ADD_AP_MOD_ARMOR")
			)
		end,
	},
	-- Ranged Attack Power - RANGED_AP, AP, AGI, INT
	{
		option = "sumRAP",
		name = StatLogic.Stats.RangedAttackPower,
		func = function(sum)
			return (GSM("MOD_RANGED_AP") + GSM("MOD_AP") - 1) * (
				sum[StatLogic.Stats.RangedAttackPower]
				+ sum[StatLogic.Stats.AttackPower]
				+ sum[StatLogic.Stats.Agility] * GSM("ADD_RANGED_AP_MOD_AGI")
				+ sum[StatLogic.Stats.Intellect] * GSM("ADD_RANGED_AP_MOD_INT")
				+ sum[StatLogic.Stats.Stamina] * GSM("ADD_AP_MOD_STA")
				+ summaryFunc[StatLogic.Stats.Armor](sum) * GSM("ADD_AP_MOD_ARMOR")
			)
		end,
	},
	-- Hit Chance - MELEE_HIT_RATING, WEAPON_SKILL
	{
		option = "sumHit",
		name = StatLogic.Stats.MeleeHit,
		func = function(sum)
			return sum[StatLogic.Stats.MeleeHit]
				+ StatLogic:GetEffectFromRating(sum[StatLogic.Stats.MeleeHitRating], StatLogic.Stats.MeleeHitRating, calcLevel)
				+ sum[StatLogic.Stats.WeaponSkill] * 0.1
		end,
		ispercent = true,
	},
	-- Hit Rating - MELEE_HIT_RATING
	{
		option = "sumHitRating",
		name = StatLogic.Stats.MeleeHitRating,
		func = function(sum)
			return sum[StatLogic.Stats.MeleeHitRating]
		end,
	},
	-- Ranged Hit Chance - MELEE_HIT_RATING, RANGED_HIT_RATING, AGI
	{
		option = "sumRangedHit",
		name = StatLogic.Stats.RangedHit,
		func = function(sum)
			return sum[StatLogic.Stats.RangedHit]
				+ StatLogic:GetEffectFromRating(sum[StatLogic.Stats.RangedHitRating], StatLogic.Stats.RangedHitRating, calcLevel)
		end,
		ispercent = true,
	},
	-- Ranged Hit Rating - RANGED_HIT_RATING
	{
		option = "sumRangedHitRating",
		name = StatLogic.Stats.RangedHitRating,
		func = function(sum)
			return sum[StatLogic.Stats.RangedHitRating]
		end,
	},
	-- Crit Chance - MELEE_CRIT, MELEE_CRIT_RATING, AGI
	{
		option = "sumCrit",
		name = StatLogic.Stats.MeleeCrit,
		func = function(sum)
			return sum[StatLogic.Stats.MeleeCrit]
				+ StatLogic:GetEffectFromRating(sum[StatLogic.Stats.MeleeCritRating], StatLogic.Stats.MeleeCritRating, calcLevel)
				+ sum[StatLogic.Stats.Agility] * StatLogic:GetCritPerAgi(class, calcLevel)
		end,
		ispercent = true,
	},
	-- Crit Rating - MELEE_CRIT_RATING
	{
		option = "sumCritRating",
		name = StatLogic.Stats.MeleeCritRating,
		func = function(sum)
			return sum[StatLogic.Stats.MeleeCritRating]
		end,
	},
	-- Ranged Crit Chance - MELEE_CRIT_RATING, RANGED_CRIT_RATING, AGI
	{
		option = "sumRangedCrit",
		name = StatLogic.Stats.RangedCrit,
		func = function(sum)
			return sum[StatLogic.Stats.RangedCrit]
				+ StatLogic:GetEffectFromRating(sum[StatLogic.Stats.RangedCritRating], StatLogic.Stats.RangedCritRating, calcLevel)
				+ sum[StatLogic.Stats.Agility] * StatLogic:GetCritPerAgi(class, calcLevel)
		end,
		ispercent = true,
	},
	-- Ranged Crit Rating - RANGED_CRIT_RATING
	{
		option = "sumRangedCritRating",
		name = StatLogic.Stats.RangedCritRating,
		func = function(sum)
			return sum[StatLogic.Stats.RangedCritRating]
		end,
	},
	-- Haste - MELEE_HASTE_RATING
	{
		option = "sumHaste",
		name = StatLogic.Stats.MeleeHaste,
		func = function(sum)
			return StatLogic:GetEffectFromRating(sum[StatLogic.Stats.MeleeHasteRating], StatLogic.Stats.MeleeHasteRating, calcLevel)
		end,
		ispercent = true,
	},
	-- Haste Rating - MELEE_HASTE_RATING
	{
		option = "sumHasteRating",
		name = StatLogic.Stats.MeleeHasteRating,
		func = function(sum)
			return sum[StatLogic.Stats.MeleeHasteRating]
		end,
	},
	-- Ranged Haste - RANGED_HASTE_RATING
	{
		option = "sumRangedHaste",
		name = StatLogic.Stats.RangedHaste,
		func = function(sum)
			return StatLogic:GetEffectFromRating(sum[StatLogic.Stats.RangedHasteRating], StatLogic.Stats.RangedHasteRating, calcLevel)
		end,
		ispercent = true,
	},
	-- Ranged Haste Rating - RANGED_HASTE_RATING
	{
		option = "sumRangedHasteRating",
		name = StatLogic.Stats.RangedHasteRating,
		func = function(sum)
			return sum[StatLogic.Stats.RangedHasteRating]
		end,
	},
	-- Expertise - EXPERTISE_RATING
	{
		option = "sumExpertise",
		name = StatLogic.Stats.Expertise,
		func = function(sum)
			return StatLogic:GetEffectFromRating(sum[StatLogic.Stats.ExpertiseRating], StatLogic.Stats.ExpertiseRating, calcLevel)
		end,
	},
	-- Expertise Rating - EXPERTISE_RATING
	{
		option = "sumExpertiseRating",
		name = StatLogic.Stats.ExpertiseRating,
		func = function(sum)
			return sum[StatLogic.Stats.ExpertiseRating]
		end,
	},
	-- Dodge Reduction - EXPERTISE_RATING, WEAPON_SKILL
	{
		option = "sumDodgeNeglect",
		name = StatLogic.Stats.DodgeReduction,
		func = function(sum)
			return floor(StatLogic:GetEffectFromRating(sum[StatLogic.Stats.ExpertiseRating], StatLogic.Stats.ExpertiseRating, calcLevel)) * 0.25
				+ sum[StatLogic.Stats.WeaponSkill] * 0.1
		end,
		ispercent = true,
	},
	-- Parry Reduction - EXPERTISE_RATING
	{
		option = "sumParryNeglect",
		name = StatLogic.Stats.ParryReduction,
		func = function(sum)
			return floor(StatLogic:GetEffectFromRating(sum[StatLogic.Stats.ExpertiseRating], StatLogic.Stats.ExpertiseRating, calcLevel)) * 0.25
		end,
		ispercent = true,
	},
	-- Weapon Average Damage - StatLogic.Stats.MinWeaponDamage, StatLogic.Stats.MaxWeaponDamage
	{
		option = "sumWeaponAverageDamage",
		name = StatLogic.Stats.AverageWeaponDamage,
		func = function(sum)
			return sum[StatLogic.Stats.MinWeaponDamage] * GSM("ADD_WEAPON_DAMAGE_AVERAGE_MOD_WEAPON_DAMAGE_MIN")
				+ sum[StatLogic.Stats.MaxWeaponDamage] * GSM("ADD_WEAPON_DAMAGE_AVERAGE_MOD_WEAPON_DAMAGE_MAX")
		end,
	},
	-- Weapon DPS - DPS
	{
		option = "sumWeaponDPS",
		name = StatLogic.Stats.WeaponDPS,
		func = function(sum)
			return sum[StatLogic.Stats.WeaponDPS]
		end,
	},
	-- Ignore Armor - IGNORE_ARMOR
	{
		option = "sumIgnoreArmor",
		name = StatLogic.Stats.IgnoreArmor,
		func = function(sum)
			return sum[StatLogic.Stats.IgnoreArmor]
		end,
	},
	-- Armor Penetration - ARMOR_PENETRATION_RATING
	{
		option = "sumArmorPenetration",
		name = StatLogic.Stats.ArmorPenetration,
		func = function(sum)
			return StatLogic:GetEffectFromRating(sum[StatLogic.Stats.ArmorPenetrationRating], StatLogic.Stats.ArmorPenetrationRating, calcLevel)
		end,
		ispercent = true,
	},
	-- Armor Penetration Rating - ARMOR_PENETRATION_RATING
	{
		option = "sumArmorPenetrationRating",
		name = StatLogic.Stats.ArmorPenetrationRating,
		func = function(sum) return
			sum[StatLogic.Stats.ArmorPenetrationRating]
		end,
	},
	------------------------------
	-- Spell Damage and Healing --
	------------------------------
	-- Spell Damage - SPELL_DMG, STA, INT, SPI
	{
		option = "sumSpellDmg",
		name = StatLogic.Stats.SpellDamage,
		func = function(sum)
			return sum[StatLogic.Stats.SpellDamage]
				+ sum[StatLogic.Stats.Strength] * GSM("ADD_SPELL_DMG_MOD_STR")
				+ sum[StatLogic.Stats.Stamina] * (GSM("ADD_SPELL_DMG_MOD_STA") + GSM("ADD_SPELL_DMG_MOD_PET_STA") * GSM("MOD_PET_STA") * GSM("ADD_PET_STA_MOD_STA"))
				+ sum[StatLogic.Stats.Intellect] * (GSM("ADD_SPELL_DMG_MOD_INT") + GSM("ADD_SPELL_DMG_MOD_PET_INT") * GSM("MOD_PET_INT") * GSM("ADD_PET_INT_MOD_INT"))
				+ sum[StatLogic.Stats.Spirit] * GSM("ADD_SPELL_DMG_MOD_SPI")
				+ summaryFunc[StatLogic.Stats.AttackPower](sum) * GSM("ADD_SPELL_DMG_MOD_AP")
		end,
	},
	-- Holy Damage - HOLY_SPELL_DMG, SPELL_DMG, INT, SPI
	{
		option = "sumHolyDmg",
		name = StatLogic.Stats.HolyDamage,
		func = function(sum)
			return GSM("MOD_SPELL_DMG") * (
				sum[StatLogic.Stats.HolyDamage]
			) + summaryFunc[StatLogic.Stats.SpellDamage](sum)
		 end,
	},
	-- Arcane Damage - ARCANE_SPELL_DMG, SPELL_DMG, INT
	{
		option = "sumArcaneDmg",
		name = StatLogic.Stats.ArcaneDamage,
		func = function(sum)
			return GSM("MOD_SPELL_DMG") * (
				sum[StatLogic.Stats.ArcaneDamage]
			) + summaryFunc[StatLogic.Stats.SpellDamage](sum)
		 end,
	},
	-- Fire Damage - FIRE_SPELL_DMG, SPELL_DMG, STA, INT
	{
		option = "sumFireDmg",
		name = StatLogic.Stats.FireDamage,
		func = function(sum)
			return GSM("MOD_SPELL_DMG") * (
				sum[StatLogic.Stats.FireDamage]
			) + summaryFunc[StatLogic.Stats.SpellDamage](sum)
		 end,
	},
	-- Nature Damage - NATURE_SPELL_DMG, SPELL_DMG, INT
	{
		option = "sumNatureDmg",
		name = StatLogic.Stats.NatureDamage,
		func = function(sum)
			return GSM("MOD_SPELL_DMG") * (
				sum[StatLogic.Stats.NatureDamage]
			) + summaryFunc[StatLogic.Stats.SpellDamage](sum)
		 end,
	},
	-- Frost Damage - FROST_SPELL_DMG, SPELL_DMG, INT
	{
		option = "sumFrostDmg",
		name = StatLogic.Stats.FrostDamage,
		func = function(sum)
			return GSM("MOD_SPELL_DMG") * (
				sum[StatLogic.Stats.FrostDamage]
			) + summaryFunc[StatLogic.Stats.SpellDamage](sum)
		 end,
	},
	-- Shadow Damage - SHADOW_SPELL_DMG, SPELL_DMG, STA, INT, SPI
	{
		option = "sumShadowDmg",
		name = StatLogic.Stats.ShadowDamage,
		func = function(sum)
			return GSM("MOD_SPELL_DMG") * (
				sum[StatLogic.Stats.ShadowDamage]
			) + summaryFunc[StatLogic.Stats.SpellDamage](sum)
		 end,
	},
	-- Healing - HEAL, AGI, STR, INT, SPI, AP
	{
		option = "sumHealing",
		name = StatLogic.Stats.HealingPower,
		func = function(sum)
			return GSM("MOD_HEALING") * (
				sum[StatLogic.Stats.HealingPower]
				+ (sum[StatLogic.Stats.Strength] * GSM("ADD_HEALING_MOD_STR"))
				+ (sum[StatLogic.Stats.Agility] * GSM("ADD_HEALING_MOD_AGI"))
				+ (sum[StatLogic.Stats.Intellect] * GSM("ADD_HEALING_MOD_INT"))
				+ (sum[StatLogic.Stats.Spirit] * GSM("ADD_HEALING_MOD_SPI"))
				+ (summaryFunc[StatLogic.Stats.AttackPower](sum) * GSM("ADD_HEALING_MOD_AP"))
			)
		end,
	},
	-- Spell Hit Chance - SPELL_HIT_RATING
	{
		option = "sumSpellHit",
		name = StatLogic.Stats.SpellHit,
		func = function(sum)
			return sum[StatLogic.Stats.SpellHit]
				+ StatLogic:GetEffectFromRating(sum[StatLogic.Stats.SpellHitRating], StatLogic.Stats.SpellHitRating, calcLevel)
		end,
		ispercent = true,
	},
	-- Spell Hit Rating - SPELL_HIT_RATING
	{
		option = "sumSpellHitRating",
		name = StatLogic.Stats.SpellHitRating,
		func = function(sum)
			return sum[StatLogic.Stats.SpellHitRating]
		end,
	},
	-- Spell Crit Chance - SPELL_CRIT_RATING, INT
	{
		option = "sumSpellCrit",
		name = StatLogic.Stats.SpellCrit,
		func = function(sum)
			return sum[StatLogic.Stats.SpellCrit]
				+ StatLogic:GetEffectFromRating(summaryFunc[StatLogic.Stats.SpellCritRating](sum), StatLogic.Stats.SpellCritRating, calcLevel)
				+ sum[StatLogic.Stats.Intellect] * StatLogic:GetSpellCritPerInt(class, calcLevel)
		end,
		ispercent = true,
	},
	-- Spell Crit Rating - SPELL_CRIT_RATING
	{
		option = "sumSpellCritRating",
		name = StatLogic.Stats.SpellCritRating,
		func = function(sum)
			return sum[StatLogic.Stats.SpellCritRating]
				+ sum[StatLogic.Stats.Spirit] * GSM("ADD_SPELL_CRIT_RATING_MOD_SPI")
		end,
	},
	-- Spell Haste - SPELL_HASTE_RATING
	{
		option = "sumSpellHaste",
		name = StatLogic.Stats.SpellHaste,
		func = function(sum)
			return StatLogic:GetEffectFromRating(sum[StatLogic.Stats.SpellHasteRating], StatLogic.Stats.SpellHasteRating, calcLevel)
		end,
		ispercent = true,
	},
	-- Spell Haste Rating - SPELL_HASTE_RATING
	{
		option = "sumSpellHasteRating",
		name = StatLogic.Stats.SpellHasteRating,
		func = function(sum)
			return sum[StatLogic.Stats.SpellHasteRating]
		end,
	},
	-- Spell Penetration - SPELLPEN
	{
		option = "sumPenetration",
		name = StatLogic.Stats.SpellPenetration,
		func = function(sum)
			return sum[StatLogic.Stats.SpellPenetration]
		end,
	},
	----------
	-- Tank --
	----------
	-- Armor - ARMOR, ARMOR_BONUS, AGI, INT
	{
		option = "sumArmor",
		name = StatLogic.Stats.Armor,
		func = function(sum)
			return GSM("MOD_ARMOR") * sum[StatLogic.Stats.Armor]
				+ sum[StatLogic.Stats.BonusArmor]
				+ sum[StatLogic.Stats.Agility] * GSM("ADD_BONUS_ARMOR_MOD_AGI")
				+ sum[StatLogic.Stats.Intellect] * GSM("ADD_BONUS_ARMOR_MOD_INT")
		 end,
	},
	-- Dodge Chance Before DR - DODGE, DODGE_RATING, DEFENSE, AGI
	{
		option = "sumDodgeBeforeDR",
		name = StatLogic.Stats.DodgeBeforeDR,
		func = function(sum)
			return sum[StatLogic.Stats.Dodge]
				+ StatLogic:GetEffectFromRating(sum[StatLogic.Stats.DodgeRating], StatLogic.Stats.DodgeRating, calcLevel)
				+ summaryFunc[StatLogic.Stats.Defense](sum) * DODGE_PARRY_BLOCK_PERCENT_PER_DEFENSE
				+ sum[StatLogic.Stats.Agility] * StatLogic:GetDodgePerAgi()
		end,
		ispercent = true,
	},
	-- Dodge Chance
	{
		option = "sumDodge",
		name = StatLogic.Stats.Dodge,
		func = function(sum, sumType)
			local dodge = summaryFunc[StatLogic.Stats.DodgeBeforeDR](sum)
			if profileDB.enableAvoidanceDiminishingReturns then
				if (sumType == "diff1") or (sumType == "diff2") then
					dodge = StatLogic:GetAvoidanceGainAfterDR(StatLogic.Stats.Dodge, dodge)
				elseif sumType == "sum" then
					dodge = StatLogic:GetAvoidanceGainAfterDR(StatLogic.Stats.Dodge, equippedDodge + dodge) - StatLogic:GetAvoidanceGainAfterDR(StatLogic.Stats.Dodge, equippedDodge)
				end
			end
			return dodge
		 end,
		ispercent = true,
	},
	-- Dodge Rating - DODGE_RATING
	{
		option = "sumDodgeRating",
		name = StatLogic.Stats.DodgeRating,
		func = function(sum)
			return sum[StatLogic.Stats.DodgeRating]
		end,
	},
	-- Parry Chance Before DR - PARRY, PARRY_RATING, DEFENSE
	{
		option = "sumParryBeforeDR",
		name = StatLogic.Stats.ParryBeforeDR,
		func = function(sum)
			return GetParryChance() > 0 and (
				sum[StatLogic.Stats.Parry]
				+ StatLogic:GetEffectFromRating(summaryFunc[StatLogic.Stats.ParryRating](sum), StatLogic.Stats.ParryRating, calcLevel)
				+ summaryFunc[StatLogic.Stats.Defense](sum) * DODGE_PARRY_BLOCK_PERCENT_PER_DEFENSE
			) or 0
		end,
		ispercent = true,
	},
	-- Parry Chance
	{
		option = "sumParry",
		name = StatLogic.Stats.Parry,
		func = function(sum, sumType)
			local parry = summaryFunc[StatLogic.Stats.ParryBeforeDR](sum)
			if profileDB.enableAvoidanceDiminishingReturns then
				if (sumType == "diff1") or (sumType == "diff2") then
					parry = StatLogic:GetAvoidanceGainAfterDR(StatLogic.Stats.Parry, parry)
				elseif sumType == "sum" then
					parry = StatLogic:GetAvoidanceGainAfterDR(StatLogic.Stats.Parry, equippedParry + parry) - StatLogic:GetAvoidanceGainAfterDR(StatLogic.Stats.Parry, equippedParry)
				end
			end
			return parry
		 end,
		ispercent = true,
	},
	-- Parry Rating - PARRY_RATING
	{
		option = "sumParryRating",
		name = StatLogic.Stats.ParryRating,
		func = function(sum)
			return sum[StatLogic.Stats.ParryRating]
				+ sum[StatLogic.Stats.Strength] * GSM("ADD_PARRY_RATING_MOD_STR")
		end,
	},
	-- Block Chance - BLOCK, BLOCK_RATING, DEFENSE
	{
		option = "sumBlock",
		name = StatLogic.Stats.BlockChance,
		func = function(sum)
			return GetBlockChance() > 0 and (
				sum[StatLogic.Stats.BlockChance]
				+ StatLogic:GetEffectFromRating(sum[StatLogic.Stats.BlockRating], StatLogic.Stats.BlockRating, calcLevel)
				+ summaryFunc[StatLogic.Stats.Defense](sum) * DODGE_PARRY_BLOCK_PERCENT_PER_DEFENSE
			) or 0
		end,
		ispercent = true,
	},
	-- Block Rating - BLOCK_RATING
	{
		option = "sumBlockRating",
		name = StatLogic.Stats.BlockRating,
		func = function(sum)
			return sum[StatLogic.Stats.BlockRating]
		end,
	},
	-- Block Value - BLOCK_VALUE, STR
	{
		option = "sumBlockValue",
		name = StatLogic.Stats.BlockValue,
		func = function(sum)
			return GetBlockChance() > 0 and (
				GSM("MOD_BLOCK_VALUE") * (
					sum[StatLogic.Stats.BlockValue]
					+ sum[StatLogic.Stats.Strength] * GSM("ADD_BLOCK_VALUE_MOD_STR")
				)
			) or 0
		end,
	},
	-- Hit Avoidance Before DR - DEFENSE
	{
		option = "sumHitAvoidBeforeDR",
		name = StatLogic.Stats.MissBeforeDR,
		func = function(sum)
			return summaryFunc[StatLogic.Stats.Defense](sum) * DODGE_PARRY_BLOCK_PERCENT_PER_DEFENSE
		end,
		ispercent = true,
	},
	-- Hit Avoidance
	{
		option = "sumHitAvoid",
		name = StatLogic.Stats.Miss,
		func = function(sum, sumType)
			local missed = summaryFunc[StatLogic.Stats.MissBeforeDR](sum)
			if profileDB.enableAvoidanceDiminishingReturns then
				if (sumType == "diff1") or (sumType == "diff2") then
					missed = StatLogic:GetAvoidanceGainAfterDR(StatLogic.Stats.Miss, missed)
				elseif sumType == "sum" then
					missed = StatLogic:GetAvoidanceGainAfterDR(StatLogic.Stats.Miss, equippedMissed + missed) - StatLogic:GetAvoidanceGainAfterDR(StatLogic.Stats.Miss, equippedMissed)
				end
			end
			return missed
		 end,
		ispercent = true,
	},
	-- Defense - DEFENSE_RATING
	{
		option = "sumDefense",
		name = StatLogic.Stats.Defense,
		func = function(sum)
			return sum[StatLogic.Stats.Defense]
				+ StatLogic:GetEffectFromRating(sum[StatLogic.Stats.DefenseRating], StatLogic.Stats.DefenseRating, calcLevel)
		end,
	},
	-- Avoidance - DODGE, PARRY, MELEE_HIT_AVOID, BLOCK(Optional)
	{
		option = "sumAvoidance",
		name = StatLogic.Stats.Avoidance,
		ispercent = true,
		func = function(sum, sumType, link)
			local dodge = summaryFunc[StatLogic.Stats.Dodge](sum, sumType, link)
			local parry = summaryFunc[StatLogic.Stats.Parry](sum, sumType, link)
			local missed = summaryFunc[StatLogic.Stats.Miss](sum, sumType, link)
			local block = 0
			if profileDB.sumAvoidWithBlock then
				block = summaryFunc[StatLogic.Stats.BlockChance](sum, sumType, link)
			end
			return parry + dodge + missed + block
		end,
	},
	-- Crit Avoidance - RESILIENCE_RATING, DEFENSE
	{
		option = "sumCritAvoid",
		name = StatLogic.Stats.CritAvoidance,
		func = function(sum)
			return StatLogic:GetEffectFromRating(sum[StatLogic.Stats.ResilienceRating], StatLogic.Stats.ResilienceRating, calcLevel)
				+ summaryFunc[StatLogic.Stats.Defense](sum) * DODGE_PARRY_BLOCK_PERCENT_PER_DEFENSE
		 end,
		ispercent = true,
	},
	-- Resilience - RESILIENCE_RATING
	{
		option = "sumResilience",
		name = StatLogic.Stats.ResilienceRating,
		func = function(sum)
			return sum[StatLogic.Stats.ResilienceRating]
		end,
	},
	-- Arcane Resistance - ARCANE_RES
	{
		option = "sumArcaneResist",
		name = StatLogic.Stats.ArcaneResistance,
		func = function(sum)
			return sum[StatLogic.Stats.ArcaneResistance]
		end,
	},
	-- Fire Resistance - FIRE_RES
	{
		option = "sumFireResist",
		name = StatLogic.Stats.FireResistance,
		func = function(sum)
			return sum[StatLogic.Stats.FireResistance]
		end,
	},
	-- Nature Resistance - NATURE_RES
	{
		option = "sumNatureResist",
		name = StatLogic.Stats.NatureResistance,
		func = function(sum)
			return sum[StatLogic.Stats.NatureResistance]
		end,
	},
	-- Frost Resistance - FROST_RES
	{
		option = "sumFrostResist",
		name = StatLogic.Stats.FrostResistance,
		func = function(sum)
			return sum[StatLogic.Stats.FrostResistance]
		end,
	},
	-- Shadow Resistance - SHADOW_RES
	{
		option = "sumShadowResist",
		name = StatLogic.Stats.ShadowResistance,
		func = function(sum)
			return sum[StatLogic.Stats.ShadowResistance]
		end,
	},
}

-- Build summaryFunc
for _, calcData in pairs(summaryCalcData) do
	summaryFunc[calcData.name] = calcData.func
end

local function sumSortAlphaComp(a, b)
	return a[1] < b[1]
end

local function WriteSummary(tooltip, output)
	if globalDB.sumBlankLine then
		tooltip:AddLine(" ")
	end
	if globalDB.sumShowTitle then
		tooltip:AddLine(HIGHLIGHT_FONT_COLOR_CODE..L["Stat Summary"]..FONT_COLOR_CODE_CLOSE)
		if globalDB.sumShowIcon then
			tooltip:AddTexture("Interface\\AddOns\\RatingBuster\\images\\Sigma")
		end
	end
	local statR, statG, statB = globalDB.sumStatColor:GetRGB()
	local valueR, valueG, valueB = globalDB.sumValueColor:GetRGB()
	for _, o in ipairs(output) do
		tooltip:AddDoubleLine(o[1], o[2], statR, statG, statB, valueR, valueG, valueB)
	end
	if globalDB.sumBlankLineAfter then
		tooltip:AddLine(" ")
	end
end

function RatingBuster:StatSummary(tooltip, link)
	-- Hide stat summary for equipped items
	if globalDB.sumIgnoreEquipped and IsEquippedItem(link) then return end

	-- Show stat summary only for highest level armor type and items you can use with uncommon quality and up
	if globalDB.sumIgnoreUnused then
		local _, _, itemRarity, _, _, _, _, _, inventoryType, _, classID, subclassID = GetItemInfo(link)

		-- Check rarity
		if not itemRarity or itemRarity < 2 then
			return
		end

		-- Check armor type
		if classID == Enum.ItemClass.Armor and armorTypes[subclassID] and (not classArmorTypes[class][subclassID]) and inventoryType ~= "INVTYPE_CLOAK" then
			return
		end

		-- Check for Red item types
		local tName = tooltip:GetName()
		if _G[tName.."TextRight3"]:GetText() and select(2, _G[tName.."TextRight3"]:GetTextColor()) < 0.2 then
			return
		end
		if _G[tName.."TextRight4"]:GetText() and select(2, _G[tName.."TextRight4"]:GetTextColor()) < 0.2 then
			return
		end
		if select(2, _G[tName.."TextLeft3"]:GetTextColor()) < 0.2 then
			return
		end
		if select(2, _G[tName.."TextLeft4"]:GetTextColor()) < 0.2 then
			return
		end
	end

	-- Ignore enchants and gems on items when calculating the stat summary
	local red = profileDB.sumGemRed.gemID
	local yellow = profileDB.sumGemYellow.gemID
	local blue = profileDB.sumGemBlue.gemID
	local meta = profileDB.sumGemMeta.gemID

	if globalDB.sumIgnoreEnchant then
		link = StatLogic:RemoveEnchant(link)
	end
	if globalDB.sumIgnoreExtraSockets then
		link = StatLogic:RemoveExtraSockets(link)
	end
	if globalDB.sumIgnoreGems then
		link = StatLogic:RemoveGem(link)
	else
		link = StatLogic:BuildGemmedTooltip(link, red, yellow, blue, meta)
	end

	-- Diff Display Style
	-- Main Tooltip: tooltipLevel = 0
	-- Compare Tooltip 1: tooltipLevel = 1
	-- Compare Tooltip 2: tooltipLevel = 2
	local id
	local tooltipLevel = 0
	local mainTooltip = tooltip
	-- Determine tooltipLevel and id
	if globalDB.calcDiff and (globalDB.sumDiffStyle == "comp") then
		-- Obtain main tooltip
		local owner = tooltip:GetOwner()
	    if owner.GetObjectType and owner:GetObjectType() == "GameTooltip" then
			mainTooltip = owner
		end
		-- Detemine tooltip level
		local _, mainlink, difflink1, difflink2 = StatLogic:GetDiffID(mainTooltip, globalDB.sumIgnoreEnchant, globalDB.sumIgnoreGems, globalDB.sumIgnoreExtraSockets, red, yellow, blue, meta)
		if link == mainlink then
			tooltipLevel = 0
		elseif link == difflink1 then
			tooltipLevel = 1
		elseif link == difflink2 then
			tooltipLevel = 2
		end
		-- Construct id
		if tooltipLevel > 0 then
			id = link..mainlink
		else
			id = "sum"..link
		end
	else
		id = StatLogic:GetDiffID(link, globalDB.sumIgnoreEnchant, globalDB.sumIgnoreGems, globalDB.sumIgnoreExtraSockets, red, yellow, blue, meta)
	end

	local numLines = tooltip:NumLines()

	-- Check Cache
	if cache[id] and cache[id].numLines == numLines then
		if table.maxn(cache[id]) == 0 then return end
		WriteSummary(tooltip, cache[id])
		return
	end

	-------------------------
	-- Build Summary Table --
	local statData = {}
	statData.sum = StatLogic:GetSum(link)
	if not statData.sum then return end
	if not globalDB.calcSum then
		statData.sum = nil
	end

	-- Ignore bags
	if not StatLogic:GetDiff(link) then return end

	-- Get Diff Data
	if globalDB.calcDiff then
		if globalDB.sumDiffStyle == "comp" then
			if tooltipLevel > 0 then
				statData.diff1 = select(tooltipLevel, StatLogic:GetDiff(mainTooltip, nil, nil, globalDB.sumIgnoreEnchant, globalDB.sumIgnoreGems, globalDB.sumIgnoreExtraSockets, red, yellow, blue, meta))
			end
		else
			statData.diff1, statData.diff2 = StatLogic:GetDiff(link, nil, nil, globalDB.sumIgnoreEnchant, globalDB.sumIgnoreGems, globalDB.sumIgnoreExtraSockets, red, yellow, blue, meta)
		end
	end
	-- Apply Base Stat Mods
	for _, v in pairs(statData) do
		v[StatLogic.Stats.Strength] = (v[StatLogic.Stats.Strength] or 0)
		v[StatLogic.Stats.Agility] = (v[StatLogic.Stats.Agility] or 0)
		v[StatLogic.Stats.Stamina] = (v[StatLogic.Stats.Stamina] or 0)
		v[StatLogic.Stats.Intellect] = (v[StatLogic.Stats.Intellect] or 0)
		v[StatLogic.Stats.Spirit] = (v[StatLogic.Stats.Spirit] or 0)
	end
	if profileDB.enableStatMods then
		for _, v in pairs(statData) do
			v[StatLogic.Stats.Strength] = v[StatLogic.Stats.Strength] * GSM("MOD_STR")
			v[StatLogic.Stats.Agility] = v[StatLogic.Stats.Agility] * GSM("MOD_AGI")
			v[StatLogic.Stats.Stamina] = v[StatLogic.Stats.Stamina] * GSM("MOD_STA")
			v[StatLogic.Stats.Intellect] = v[StatLogic.Stats.Intellect] * GSM("MOD_INT")
			v[StatLogic.Stats.Spirit] = v[StatLogic.Stats.Spirit] * GSM("MOD_SPI")
		end
	end

	local summary = {}
	for _, calcData in pairs(summaryCalcData) do
		if profileDB[calcData.option] then
			local entry = {
				name = calcData.name,
				ispercent = calcData.ispercent,
			}
			for statDataType, statTable in pairs(statData) do
				entry[statDataType] = calcData.func(statTable, statDataType, link)
			end
			tinsert(summary, entry)
		end
	end

	local calcSum = globalDB.calcSum
	local calcDiff = globalDB.calcDiff

	local showZeroValueStat = profileDB.showZeroValueStat
	------------------------
	-- Build Output Table --
	local output = {}
	output.numLines = numLines
	for _, t in ipairs(summary) do
		local n, s, d1, d2, ispercent = t.name, t.sum, t.diff1, t.diff2, t.ispercent
		local right, left
		local skip
		if not showZeroValueStat then
			if (s == 0 or not s) and (d1 == 0 or not d1) and (d2 == 0 or not d2) then
				skip = true
			end
		end
		if not skip then
			if calcSum and calcDiff then
				local d = ((not s) or ((s - floor(s)) == 0)) and ((not d1) or ((d1 - floor(d1)) == 0)) and ((not d2) or ((d2 - floor(d2)) == 0)) and not ispercent
				if s then
					if d then
						s = ("%d"):format(s)
					elseif ispercent then
						s = ("%.2f%%"):format(s)
					else
						s = ("%.1f"):format(s)
					end
					if d1 then
						if d then
							d1 = colorNum(("%+d"):format(d1), d1)
						elseif ispercent then
							d1 = colorNum(("%+.2f%%"):format(d1), d1)
						else
							d1 = colorNum(("%+.1f"):format(d1), d1)
						end
						if d2 then
							if d then
								d2 = colorNum(("%+d"):format(d2), d2)
							elseif ispercent then
								d2 = colorNum(("%+.2f%%"):format(d2), d2)
							else
								d2 = colorNum(("%+.1f"):format(d2), d2)
							end
							right = ("%s (%s||%s)"):format(s, d1, d2)
						else
							right = ("%s (%s)"):format(s, d1)
						end
					else
						right = s
					end
				else
					if d1 then
						if d then
							d1 = colorNum(("%+d"):format(d1), d1)
						elseif ispercent then
							d1 = colorNum(("%+.2f%%"):format(d1), d1)
						else
							d1 = colorNum(("%+.1f"):format(d1), d1)
						end
						if d2 then
							if d then
								d2 = colorNum(("%+d"):format(d2), d2)
							elseif ispercent then
								d2 = colorNum(("%+.2f%%"):format(d2), d2)
							else
								d2 = colorNum(("%+.1f"):format(d2), d2)
							end
							right = ("(%s||%s)"):format(d1, d2)
						else
							right = ("(%s)"):format(d1)
						end
					end
				end
			elseif calcSum then
				if s then
					if (s - floor(s)) == 0 then
						s = ("%d"):format(s)
					elseif ispercent then
						s = ("%.2f%%"):format(s)
					else
						s = ("%.1f"):format(s)
					end
					right = s
				end
			elseif calcDiff then
				local d = ((not d1) or (d1 - floor(d1)) == 0) and ((not d2) or ((d2 - floor(d2)) == 0))
				if d1 then
					if d then
						d1 = colorNum(("%+d"):format(d1), d1)
					elseif ispercent then
						d1 = colorNum(("%+.2f%%"):format(d1), d1)
					else
						d1 = colorNum(("%+.1f"):format(d1), d1)
					end
					if d2 then
						if d then
							d2 = colorNum(("%+d"):format(d2), d2)
						elseif ispercent then
							d2 = colorNum(("%+.2f%%"):format(d2), d2)
						else
							d2 = colorNum(("%+.1f"):format(d2), d2)
						end
						right = ("%s||%s"):format(d1, d2)
					else
						right = d1
					end
				end
			end
			if right then
				left = L[n]
				tinsert(output, {left, right})
			end
		end
	end
	-- sort alphabetically if option enabled
	if globalDB.sumSortAlpha then
		tsort(output, sumSortAlphaComp)
	end
	-- Write cache
	cache[id] = output
	if table.maxn(output) == 0 then return end
	WriteSummary(tooltip, output)
end

function RatingBuster:PerformanceProfile()
	if not GetCVarBool("scriptProfile") then
		print("The console variable \"scriptProfile\" must be enabled to do performance profiling.\nEnable it and reload to continue. Be sure to disable it again when you are finished.")
		return
	end

	-- Process the tooltips for all of the player's equipped gear
	for i = INVSLOT_FIRST_EQUIPPED, INVSLOT_LAST_EQUIPPED do
		GameTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
		GameTooltip:SetInventoryItem("player", i)
		self.ProcessTooltip(GameTooltip)
		GameTooltip:Hide()
	end

	local tables = {
		["RatingBuster"] = self,
		["StatLogic"] = StatLogic
	}

	for name, addonTable in pairs(tables) do
		print(name)
		print("(ms) (count) (function)")
		local unsorted = {}
		local rankings = {}
		for k,v in pairs(addonTable) do
			if type(v) == "function" then
				local time, count = GetFunctionCPUUsage(v, false)
				if count > 0 then
					unsorted[time] = {
						count = count,
						name = k
					}
					tinsert(rankings, time)
				end
			end
		end
		table.sort(rankings, function(a,b) return a > b end)
		for i, time in ipairs(rankings) do
			if i <= 10 then
				print(string.format("  %2.2f %4d %s", time, unsorted[time].count, unsorted[time].name))
			end
		end
	end
end